# FILE: mmdisp.tcl
#
# Main Tcl/Tk script for OOMMF 2D display program
#
# Requires Tcl/Tk script files: misc.tcl dialog.tcl
#  They must be source'd before this one, in that order.
#

# Requirements:
#
# The HelpBrowser and Url packages support the Help menu
package require HelpBrowser 0
package require Url 0

set mmd_log [em_extension ApplicationLog]

# If the --server option appeared the command line, start a server.
if {$mmd_serverflag} {
    package require net 0
    #set mmd_serverlog [em_log new -title "mmDisp server log"]
    set mmd_protocol [net_protocol new -name [list OOMMF display protocol 0.0]]
    $mmd_protocol AddMessage start bye {} {
        return [list close [list 0 Bye!]]
    }
    $mmd_protocol AddMessage start openfile { filename } {
	if [string match "EmptyMesh" [GetMeshType]] {
	    # If no file currently displayed, use auto-display
	    set result [ReadAndDeleteFile $filename]
	} else {
	    # Otherwise, keep all current size and display parameters
	    global ViewportWidth ViewportHeight DisplayRotation
	    set fixdisplay(subsample) 1
	    set fixdisplay(arrowmag) 1
	    set fixdisplay(zoom) 1
	    set result [ReadAndDeleteFile $filename \
		    $ViewportWidth $ViewportHeight \
		    $DisplayRotation fixdisplay]
	}
	return [list start [list 0 $result]]
    }
    $mmd_protocol AddMessage start HasGui {} {
        return [list start [list 0 0]]
    }
    #set mmd_server [net_server new -protocol $mmd_protocol \
    #        -log $mmd_serverlog -register 1]
    set mmd_server [net_server new -protocol $mmd_protocol \
            -log [em_extension ApplicationLog] -register 1 -alias mmDisp]
    # Start server on an OS-selected port.
    $mmd_server Start 0

    # Set up to be sure that server shuts down on exit.  
    rename exit tcl_exit
    proc exit {{code 0}} {
        global mmd_server
        if {[info exists mmd_server]} {
            catch { $mmd_server delete }  ;# Safety catch
        }
        # Do something smarter eventually.  For now, 2 seconds should
        # be enough time for the server to shut down cleanly.
        after 2000 tcl_exit $code
    }
}

proc ReadAndDeleteFile { filename {width {}} {height {}} {rot {}} \
	{fixdisplayName {}}} {
    set fixname {}
    if { ![string match {} $fixdisplayName] } {
	upvar $fixdisplayName fixdisplay
	if {[info exists fixdisplay]} { set fixname fixdisplay }
    }
    ReadFile $filename $width $height $rot $fixname
    file delete $filename
}

# Allow loading of other support procs
lappend auto_path [file join \
                   $omfmmd_library [OMFMMD_FileNameBase][OMFMMD_Version]]

source [file join $omfmmd_library [OMFMMD_FileNameBase][OMFMMD_Version] \
                  misc.tcl]

########################### PROGRAM DOCUMENTATION ###################
set ProgName "mmDisp"   ;# Don't use any spaces in this, because
                         # it may be used for naming variables
wm title . $ProgName<[pid]>  ;# Window title to display while program is
                      ## initializing.
set VersNumber 0.0b0
set VersDate "17-Sep-1997"
set ProgAuthor "Michael J. Donahue"
set ProgContact "michael.donahue@nist.gov"
proc AboutInfo { menuname itemlabel } {
    global ProgName ProgAuthor ProgContact
    global VersNumber VersDate omfTclBuildVersion
    set aboutwin [OMF_InfoDialog \
            "$ProgName-v$VersNumber\n$VersDate\n\n\
	     Built against Tcl $omfTclBuildVersion\n\
	     Running on Tcl [info patchlevel]\n\n\
	     Send bug reports to: $ProgContact\n" \
	    "About $ProgName" none 12c center]
    bind $aboutwin <Destroy> "+
        # Ignore child window <Destroy> events
	if \[string compare $aboutwin %W\] continue
        $menuname entryconfigure $itemlabel -state normal
    "
    $menuname entryconfigure $itemlabel -state disabled
}
proc CreditsInfo { menuname itemlabel } {
    global ProgName ProgAuthor ProgContact
    set creditwin [ OMF_InfoDialog  \
            "$ProgName\n$ProgAuthor\n$ProgContact\n\n\
	    Help browser, installation and environment tools support\n\
	    Donald G. Porter\ndonald.porter@nist.gov\n\n\
	    User's manual and help pages\n\
	    Jason Eicke\njeicke@seas.gwu.edu\n" \
	    "$ProgName Credits" none 12c center]
    bind $creditwin <Destroy> "+
        # Ignore child window <Destroy> events
	if \[string compare $creditwin %W\] continue
        $menuname entryconfigure $itemlabel -state normal
    "
    $menuname entryconfigure $itemlabel -state disabled
}
proc LicenseInfo { menuname itemlabel } {
    global ProgName
    set licensewin [OMF_InfoDialog \
	    "This software was produced by NIST, an agency of the \
	    U.S. government, and is by statute not subject to \
	    copyright in the United States.\n\nRecipients of this \
	    software assume all responsibilities associated with \
	    its operation, modification, and maintenance." \
	    "$ProgName License" none 12c]
    bind $licensewin <Destroy> "+
        # Ignore child window <Destroy> events
	if \[string compare $licensewin %W\] continue
        $menuname entryconfigure $itemlabel -state normal
    "
    $menuname entryconfigure $itemlabel -state disabled
}


########################### GLOBAL VARIABLES ########################
if {![info exists DEBUGLVL]} { set DEBUGLVL 0 } ;# Value of C++ compile-
                      ## time macro DEBUGLVL.  Should be set inside
                      ## Oommf_Init() (currently in OOMMF/oomf/oomfInit.cc)
set DrawFrameCount  0 ;# This variable is incremented by DrawFrame each
                      ## time it is called, so this variable can be used
                      ## for profiling.
set UpdateViewCount 0 ;# Any time the screen display is modified, this
                      ## value will be incremented.  A trace on this
                      ## variable can be used by any routine (e.g. the
                      ## print dialog) that needs to take action upon
                      ## such events.  Currently this variable is written
                      ## to only by the PackViewport proc.  NOTE: Scrolling
                      ## changes are NOT tracked by this variable, at this
                      ## time. (-mjd 97/2/11)
set canvas .canvas_frame.canvas   ;# Name of canvas widget

set DefaultWindowWidth   640
set DefaultWindowHeight  480
set DefaultCellSize 25
# Note: DefaultCellSize must be > 0
set DefaultViewportWidth  $DefaultWindowWidth  ;# Safety.  These should be
set DefaultViewportHeight $DefaultWindowHeight ;# set during Init phase.
set MinViewportWidth  0
set MinViewportHeight 0
set ViewportWidth     0
set ViewportHeight    0
# NOTE: There are 3 different width/height pairs, and at any time
#   they may have completely different values.  ViewportWidth/Height
#   is requested viewport width/height, *including* scrollbars as needed.
#   The requested value will be modified as necessary to meet the
#   MinViewportWidth/Height restrictions.  The actual size of the viewing
#   area, *not* including the scrollbars, can be obtained by querying
#   the canvas widget about its "-width" & "-height".  The whole canvas
#   size can be obtained from the canvas "-scrollregion" configuration
#   option, or preferably from the "frame" C++ class by calling GetFrameSize.
#
set FrameScale        0
set FrameArrowMag     1.0
set DisplayRotation   0
set SubsampleRate     0
set SubsampleXOffset  0    ;# Currently (1-Feb-97) not used.
set SubsampleYOffset  0    ;# Currently (1-Feb-97) not used.
set FileId           0     ;# Incremented each time a new file is loaded.
                           ## Traces may be set on this variable to update
                           ## file specific quantities, e.g., filename,
                           ## file description, mesh type, etc.
set FrameSubsampleIncrement 0.1  ;# Subsample slider increment size
set ColorMap(base) black   ;# Color used for in-plane arrows
set ColorMap(in)   red     ;# Color used for arrows pointing into-page
set ColorMap(out)  blue    ;# Color used for arrows pointing out-of-page

# Misc. constants
set SmallZoom 1.1486984  ;# Amt to enlarge on Page_Up (Prior). This is 2^(1/5)
set LargeZoom 2          ;# Amount to enlarge on Shift-Page_Up

# These scroll state variables should be modified *only* in
# proc PackViewport
set XScrollState      0
set YScrollState      0

# Whether or not the control bar is displayed
set omfCtrlBarState 1   ;# 1->on, 0->off

# Display state recover stack
# (Currently only 1 entry deep, but may be made deeper in future
set omfSavedDisplayState(zoom)     $FrameScale
set omfSavedDisplayState(xcenter)  0.5
set omfSavedDisplayState(ycenter)  0.5
set omfSavedDisplayState(rot)      $DisplayRotation
set omfSavedDisplayState(arrowmag) $FrameArrowMag
proc OMF_SaveDisplayState { {zoom ""} {xcenter ""}  {ycenter ""} {rot ""} \
	{arrowmag ""}} {
    global omfSavedDisplayState FrameScale DisplayRotation FrameArrowMag
    if {"$zoom"==""} { set zoom $FrameScale }
    if {"$xcenter"=="" || "$ycenter"==""} {
        GetScrollCenter curx cury
        if {"$xcenter"==""} { set xcenter $curx }
        if {"$ycenter"==""} { set ycenter $cury }
    }
    if {"$rot"==""} { set rot $DisplayRotation }
    if {"$arrowmag"==""} { set arrowmag $FrameArrowMag }
    set omfSavedDisplayState(zoom) $zoom
    set omfSavedDisplayState(xcenter)  $xcenter
    set omfSavedDisplayState(ycenter)  $ycenter
    set omfSavedDisplayState(rot)      $rot
    set omfSavedDisplayState(arrowmag) $arrowmag
}
proc OMF_RecoverDisplayState {} {
    global omfSavedDisplayState DisplayRotation
    # Note: SetScale & SetArrowMag call OMF_SaveDisplayState
    #   & thereby reset these values
    set oldzoom     $omfSavedDisplayState(zoom)
    set oldxcenter  $omfSavedDisplayState(xcenter)
    set oldycenter  $omfSavedDisplayState(ycenter)
    set oldrot      $omfSavedDisplayState(rot)
    set oldarrowmag $omfSavedDisplayState(arrowmag)
    # Make rotation adjustments to xcenter & ycenter
    set dummywidth 0; set dummyheight 0
    OMF_RotateDimensions [expr $DisplayRotation-$oldrot] \
            dummywidth dummyheight oldxcenter oldycenter
    # Undo
    SetArrowMag $oldarrowmag
    SetScale $oldzoom
    SetScrollCenter $oldxcenter $oldycenter
} 

##### Window geometry requirements #####################################
set RootHeadHeight 0           ;# Height of menubar + action bar
set ScrollCrossDim 0           ;# Width of yscroll, height of xscroll

proc OMF_SetGeometryRequirements {} {
    global MinViewportWidth MinViewportHeight RootHeadHeight
    global omfCtrlBarState

    if {$omfCtrlBarState} {
	WaitVisible .ctrlbar 500
	set size [expr [winfo reqheight .ctrlbar.subsample]\
		+[winfo reqheight .ctrlbar.framescale]]
	.ctrlbar.rotation configure -height $size -width $size
    }

    set MinViewportHeight 0
    set MinViewportWidth [winfo reqwidth .menubar]
    set RootHeadHeight [winfo reqheight .menubar]
    if {$omfCtrlBarState} {
        set RootHeadHeight [expr $RootHeadHeight \
                + [winfo reqheight .ctrlbar]]
        set temp [winfo reqwidth .ctrlbar]
        if {$temp>$MinViewportWidth} {
            set MinViewportWidth $temp
        }
    }
    set minwidth $MinViewportWidth
    set minheight [expr $RootHeadHeight + $MinViewportHeight]
    wm minsize . $minwidth $minheight
    wm resizable . 1 1     ;# Safety.  This is supposed to be the default.
}


### Printing variable array omfPrint(): ################################
#    canvas     -   name of canvas widget
#    inname     -   name of file being displayed
#    outname    -   name of output (postscript) file or pipe
#    orient     -   portrait | landscape
#    size       -   output width *or* height (not both), in the format
#                     '-pagewidth 3.i'  (3 inches width), or
#                     '-pageheight 5.c' (5 cm high), for example
#    papertype  -   letter | legal | A4 | A3
#    crop       -    1 | 0  (1=>print only displayed region of canvas)
#    xmargin    -    left/right margin, with units.  E.g.: 1.i
#    ymargin    -    top/bottom margin,  as above
#    position   -    One of: c nw n ne e se s sw w
#    done       -    Written to after above values are set as desired.
#                    Should be one of "OK" or "Cancel".  The standard
#                    print trace proc (PrintDialogTrace) will actually
#                    print iff omfPrint(done) == "OK"
set omfPrint(canvas)    ".canvas_frame.canvas"
set omfPrint(inname)    ""
set omfPrint(outname)   " | lpr"
set omfPrint(orient)    "landscape"
set omfPrint(size)      "-pagewidth 6i"
set omfPrint(papertype) "letter"
set omfPrint(crop)      1
set omfPrint(xmargin)   "1i"
set omfPrint(ymargin)   "1i"
set omfPrint(position)  "c"
set omfPrint(done)      "Cancel"

if { ![info exists omfPlatform] } {
    $mmd_log Append "ERROR IN MMDISP.TCL: omfPlatform not defined!"
} else { 
    if { [string compare $omfPlatform windows] == 0 } {
	set omfPrint(outname)  {}
    }
}

##### Combined interface to mesh object ################################
proc OMF_MeshInfo { command option } {
    # Currently only supports "cget" command, but in the future
    # "configure" should be added.
    if { ![string match "cget" "$command"] } {
	OMF_NonFatalError "Error in OMF_MeshInfo: Invalid command $cmd"
	return
    }
    set value {}
    switch -exact -- $option {
	-filename { set value [GetMeshName]  }
	-title    { set value [GetMeshTitle] }
	-gridtype { set value [GetMeshType]  }
	-description { set value [GetMeshDescription] }
	default   {
	    OMF_NonFatalError "Error in OMF_MeshInfo: Invalid option $option"
	    return
	}
    }
    return $value
}

########################## DEBUGGING PROC'S ############################
proc PrintDimensions { window } {
    set width [$window cget -width]
    if [catch { set height [$window cget -height] }] then {
        set height "-1"
    }
    global mmd_log
    $mmd_log Append [format "%25s  Width=%3d, Height=%3d" $window $width \
            $height]
}

proc PrintSizes {} {
    global mmd_log
    global MinViewportWidth MinViewportHeight ViewportWidth ViewportHeight
    $mmd_log Append "MinViewportWidth/Height: $MinViewportWidth/$MinViewportHeight" 
    $mmd_log Append "   ViewportWidth/Height: $ViewportWidth/$ViewportHeight" 
    GetRootDimensions width height
    $mmd_log Append [format "%25s  Width=%3d, Height=%3d" . $width $height]
    PrintDimensions .canvas_frame.canvas
    PrintDimensions .canvas_frame.xscroll
    PrintDimensions .canvas_frame.yscroll
    $mmd_log Append [.canvas_frame.canvas configure -scrollregion]
}

proc PrintDrawFrameCount { name elt op } {
    global DrawFrameCount mmd_log
    $mmd_log Append "DrawFrame call count: $DrawFrameCount"
}
trace variable DrawFrameCount w PrintDrawFrameCount

proc PrintUpdateViewCount { name elt op } {
    global UpdateViewCount mmd_log
    $mmd_log Append "UpdateView call count: $UpdateViewCount"
}
trace variable UpdateViewCount w PrintUpdateViewCount

########################## DISPLAY PROC'S ###########################
proc UpdateWindowTitle {name elt op} {
    global ProgName
    set title [file tail [GetMeshTitle]]
    if {"$title"==""} { 
        set title "$ProgName<[pid]>"
    } else {
        set title "$ProgName<[pid]>: $title"
    }
    wm title . $title
    wm iconname . $title
}
trace variable FileId w UpdateWindowTitle

proc UpdateUserVariables {} {
    global DisplayRotation SubsampleRate FrameScale FrameArrowMag
    global FrameSubsampleIncrement
    GetFrameRotation DisplayRotation
    GetFrameSubsampling temp   ; set SubsampleRate [OMF_FormatNumber $temp]
    GetFrameArrowMag    temp   ; set FrameArrowMag [OMF_FormatNumber $temp]
    GetFrameScale       temp   ; set FrameScale    [OMF_FormatNumber $temp]
    set newincr [GetMeshIncrement]
    if {$newincr>0} { set FrameSubsampleIncrement $newincr }
}

proc UpdateWidgetGeometries {} {
# Calls the root WindowManager to update all widget sizes
    # We need to explicitly call "update" so we can disable
    # call to ResizeViewport from <Configure> binding.
    #   First, remove <Configure> binding
    set confbind [bind .canvas_frame.canvas <Configure>]
    bind .canvas_frame.canvas <Configure> ""
    #   Now update window geometries
    global ViewportWidth ViewportHeight MinViewportWidth MinViewportHeight
    global RootHeadHeight
    if { $ViewportWidth < $MinViewportWidth } {
        set width $MinViewportWidth
    } else {
        set width $ViewportWidth
    }
    if { $ViewportHeight < $MinViewportHeight } {
        set height $MinViewportHeight
    } else {
        set height $ViewportHeight
    }
    set height [expr $height + $RootHeadHeight]
    update idletasks  ;# Stops redraw "bouncing" during "wm geometry" call
#    wm geometry . "${width}x${height}"
# Hmm, the above doesn't seem to work.
    wm geometry . ""
    #  Then rebind canvas resize proc
    bind .canvas_frame.canvas <Configure> "$confbind"
}

proc GetScrollCenter { xcenter_name ycenter_name } {
    upvar $xcenter_name xcenter
    upvar $ycenter_name ycenter
    set xscrpos "[.canvas_frame.canvas xview]"
    set yscrpos "[.canvas_frame.canvas yview]"
    set xcenter [expr ([lindex $xscrpos 0]+[lindex $xscrpos 1])/2.]
    set ycenter [expr ([lindex $yscrpos 0]+[lindex $yscrpos 1])/2.]
}

proc SetScrollCenter { xcenter ycenter {xscrwidth ""} {yscrwidth ""}} {
    # All imports should be in range [0,1]
    if {"$xscrwidth" == ""} {
        set xscrpos "[.canvas_frame.canvas xview]"
        set xscrwidth [expr [lindex $xscrpos 1] - [lindex $xscrpos 0]]
    }
    if {"$yscrwidth" == ""} {
        set yscrpos "[.canvas_frame.canvas yview]"
        set yscrwidth [expr [lindex $yscrpos 1] - [lindex $yscrpos 0]]
    }
    set xstart [expr $xcenter - $xscrwidth/2.]
    set ystart [expr $ycenter - $yscrwidth/2.]
    if {$xstart<0} {set xstart 0} elseif {$xstart>1} {set xstart 1}
    if {$ystart<0} {set ystart 0} elseif {$ystart>1} {set ystart 1}
    .canvas_frame.canvas xview moveto $xstart
    .canvas_frame.canvas yview moveto $ystart
}

proc DetermineScrollStates { viewwidth viewheight framewidth frameheight \
        newxstatename newystatename } {
    # Determines which scrollbars will need to be activated, given a
    # canvas size of framewidth x frameheight, a viewport size of
    # viewwidth x viewheight, and a scrollbar cross dimension of
    # ScrollCrossDim.  States (0=>off, 1=>on) are returned in
    # newxstatename and newystatename.
    # Note: This code does *not* take MinViewportWidth/Height into account.
    global ScrollCrossDim
    upvar $newxstatename newxstate
    upvar $newystatename newystate
    set minheight [expr $viewheight - $ScrollCrossDim]
    set minwidth  [expr $viewwidth  - $ScrollCrossDim]
    if {($framewidth<=$viewwidth) && ($frameheight<=$viewheight)} {
        set newxstate 0; set newystate 0;
    } elseif {($framewidth>$viewwidth) && ($frameheight<=$minheight)} {
        set newxstate 1; set newystate 0;
    } elseif {($framewidth<=$minwidth) && ($frameheight>$viewheight)} {
        set newxstate 0; set newystate 1;
    } else {
        set newxstate 1; set newystate 1;
    }
}

proc PackViewport { {newviewwidth ""} {newviewheight ""} {doUpdate 1}} {
    # Set doUpdate to 0 if you want to do the update and
    # UpdateWidgetGeometries in the calling routine.
    global ViewportWidth ViewportHeight MinViewportWidth MinViewportHeight
    global XScrollState YScrollState
    global ScrollCrossDim

    OMF_PushCursor  ;# Save current cursor & display "watch" cursor
    # Wrap following in a catch, to insure OMF_PopCursor gets called.
    if [catch {
	if {$doUpdate} OMF_SoftUpdate

	# Set default values
	if { "$newviewwidth" == "" } {
	    set newviewwidth $ViewportWidth
	} else {
	    set ViewportWidth $newviewwidth
	}
	if { "$newviewheight" == "" } {
	    set newviewheight $ViewportHeight
	} else {
	    set ViewportHeight $newviewheight
	}

	# Determine real viewport width & height
	if { $newviewwidth < $MinViewportWidth } {
	    set vwidth $MinViewportWidth
	} else {
	    set vwidth $newviewwidth
	}
	if { $newviewheight < $MinViewportHeight } {
	    set vheight $MinViewportHeight
	} else {
	    set vheight $newviewheight
	}

	# Determine new scroll states & canvas display size
	GetFrameSize fwidth fheight
	DetermineScrollStates $vwidth $vheight $fwidth $fheight \
		newxstate newystate
	if { $newystate } {
	    set cwidth [expr $vwidth - $ScrollCrossDim]
	} else {
	    set cwidth $vwidth
	}
	if { $newxstate } { 
	    set cheight [expr $vheight - $ScrollCrossDim]
	} else {
	    set cheight $vheight
	}

	# Set new scrollregion
	.canvas_frame.canvas configure \
		-scrollregion "0 0 $fwidth $fheight" \
		-width $cwidth -height $cheight

	# Everything is now sized;  If scrollstatus has changed, we
	# need to repack display
	if { ($newxstate != $XScrollState) \
		|| ($newystate != $YScrollState) } {
	    # Unpack as needed
	    if { $XScrollState && !$newxstate} {
		pack forget .canvas_frame.xscroll
	    }
	    if { $YScrollState } { 
		# If $YScrollState is true, then y-scroll needs unpacking:
		# At least one of x-scroll or y-scroll is changing state,
		# and either way y-scroll packing changes.
		pack forget .canvas_frame.yscroll
		if { $XScrollState } {
		    pack forget .canvas_frame.yscrollbox
		}
	    }
	    # Now pack
	    if { !$XScrollState && $newxstate } {
		pack .canvas_frame.xscroll -side bottom -fill x \
			-before .canvas_frame.canvas
	    }
	    if { $newystate } {
		# If $newystate is true, then y-scroll needs packing:
		# At least one of x-scroll or y-scroll is changing state,
		# and either way y-scroll packing changes.
		if { $newxstate } then { 
		    # Both x & y scrollbars are displayed; Use
		    # corner box in lower righthand corner.
		    pack .canvas_frame.yscroll -side right -fill y \
			    -in .canvas_frame.yscrollbox
		    pack .canvas_frame.yscrollbox -side right -fill y \
			    -before .canvas_frame.xscroll
		} else {
		    pack .canvas_frame.yscroll -side right -fill y \
			    -before .canvas_frame.canvas
		}
	    }
	    set XScrollState $newxstate
	    set YScrollState $newystate
	}
	if {$doUpdate} UpdateWidgetGeometries
    } errmsg ] {
	OMF_NonFatalError "Error in PackViewport: $errmsg"
    }
    OMF_PopCursor ;# Replace original cursor
    # Notify any interested routine of display change:
    global UpdateViewCount ; incr UpdateViewCount
    if {$doUpdate} { update idletasks }
}

###################### ACTION PROC'S ###############################
# ResizeViewport: Change viewport size.  The imports are the
# requested canvas size (-width & -height) *with the current
# scrollbar status*!!!  To get the true viewport request sizes,
# you need to add on the scrollbar cross dimensions for those
# scrollbars that are active.
# Note: This procedure is bound to .canvas_frame.canvas <Configure>.
# BE CAREFUL! It may get bound & unbound at several places.
proc ResizeViewport { width height } {
    global XScrollState YScrollState ScrollCrossDim
    if { $YScrollState } {
        set width [expr $width + $ScrollCrossDim]
    }
    if { $XScrollState } {
        set height [expr $height + $ScrollCrossDim]
    }
    PackViewport $width $height
}

proc DisplayRotationTrace { name elt op } {
    global DisplayRotation ViewportHeight ViewportWidth
    GetFrameRotation old_rotation
    set angle [expr $DisplayRotation - $old_rotation]
    # Determine rotation step: 0, 1, 2 or 3
    set irotstep [expr round($angle/90.)%4]
    if { $irotstep == 0 } then {
        # No change
        GetFrameRotation DisplayRotation
        return
    }

    # Disable canvas <Configure> binding to protect against
    # automated resize events
    set confbind [bind .canvas_frame.canvas <Configure>]
    bind .canvas_frame.canvas <Configure> ""

    # Get current scroll positions
    GetScrollCenter xscrcenter yscrcenter

    # Round new rotation direction to nearest multiple of 90
    set rot [expr 90*round($DisplayRotation/90.)]

    # Adjust to range [0,360)
    set rot [expr round(fmod($rot,360.))]
    if { $rot < 0 } { incr rot 360 }

    # Rotate display
    SetFrameRotation $rot
    GetFrameRotation DisplayRotation
    UpdateCoordDisplay $rot

    # Addjust viewport width, height & scroll positions
    # Keep track of scrollbar centers rather than ends to
    # better handle case where displayed viewport width-height
    # values change due to MinViewportWidth/Height restrictions.
    set new_xscrcenter $xscrcenter
    set new_yscrcenter $yscrcenter
    set new_width      $ViewportWidth
    set new_height     $ViewportHeight
    OMF_RotateDimensions $angle new_width new_height \
            new_xscrcenter new_yscrcenter

    DrawFrame .canvas_frame.canvas
    PackViewport $new_width $new_height 1
    # We apparently need 'UpdateWidgetGeometries' ('1' above)
    # before SetScrollCenter in the case where MinViewportWidth
    # toggles scrollbars on & off.
    SetScrollCenter $new_xscrcenter $new_yscrcenter
    bind .canvas_frame.canvas <Configure> "$confbind"
}

proc SetScale { scale } {
    global FrameScale
    if { $scale == $FrameScale } { return }
    GetScrollCenter xcenter ycenter
    if { $scale > 0 } {
        OMF_SaveDisplayState $FrameScale $xcenter $ycenter
        SetFrameScale $scale
    }
    GetFrameScale scale
    set FrameScale [OMF_FormatNumber $scale]
    DrawFrame .canvas_frame.canvas
    PackViewport "" "" 0
    # UpdateWidgetGeometries #I think this isn't needed anymore (-mjd 97/2/12)
    SetScrollCenter $xcenter $ycenter
    update idletasks
}

proc SetArrowMag { arrowmag } {
    global FrameArrowMag
    if { ($arrowmag == $FrameArrowMag) || ($arrowmag<0) } { return }
    SetFrameArrowMag $arrowmag
    GetFrameArrowMag arrowmag
    set FrameArrowMag [OMF_FormatNumber $arrowmag]
    DrawFrame .canvas_frame.canvas
    update idletasks
}

proc SetArrowMagTrace { name elt op } {
    global FrameArrowMag FrameArrowMagOld
    if { ![info exists FrameArrowMagOld] } {
	set FrameArrowMagOld [expr $FrameArrowMag - 1]  ;# Force update
    }
    set request $FrameArrowMag
    set FrameArrowMag $FrameArrowMagOld
    SetArrowMag $request
    set FrameArrowMagOld $FrameArrowMag
}

proc SetSubsamplingRate { newrate } {
    # Note: This display will be updated only if SubsampleRate
    #       gets written to.  Since Subsample rate has several a
    #       write traces on it, we delay writing as much as possible.
    global SubsampleRate
    if { $newrate<0 || $newrate == $SubsampleRate } return
    OMF_SoftUpdate
    SetFrameSubsampling $newrate
    GetFrameSubsampling newrate
    if { $newrate != $SubsampleRate } {
	set SubsampleRate [OMF_FormatNumber $newrate]
	DrawFrame .canvas_frame.canvas
        PackViewport
    }
    return
}

proc WrapDisplay { } {
# Set viewport size to framesize
    GetFrameSize w h
    PackViewport $w $h
}

proc FillDisplay {} {
# Set framesize to viewport size
    global canvas
    global ViewportWidth ViewportHeight
    global MinViewportWidth MinViewportHeight
    OMF_SaveDisplayState
    set w $ViewportWidth
    if { $w < $MinViewportWidth } { set w $MinViewportWidth }
    set h $ViewportHeight
    if { $h < $MinViewportHeight } { set h $MinViewportHeight }
    SetFrameSize $w $h
    global FrameScale
    GetFrameScale temp  ;  set FrameScale [OMF_FormatNumber $temp]
    PackViewport "" "" 0
    $canvas xview moveto 0;  $canvas yview moveto 0
    DrawFrame .canvas_frame.canvas
    update idletasks
}

proc ClearFrame {} {
    global ViewportWidth ViewportHeight DisplayRotation
    set fixdisplay(subsample) 1
    set fixdisplay(arrowmag) 1
    set fixdisplay(zoom) 1
    ReadFile {} $ViewportWidth $ViewportHeight \
	    $DisplayRotation fixdisplay
}

proc ReadFile { name {width {}} {height {}} {rot {}} {fixdisplayName {}}} {
    global FileId canvas

    global ViewportWidth ViewportHeight DisplayRotation
    global DefaultViewportWidth DefaultViewportHeight

    if {[string match {} $width]} {
	set width $DefaultViewportWidth
    }
    if {[string match {} $height]} {
	set height $DefaultViewportHeight
    }
    if {[string match {} $rot]} {
	set rot $DisplayRotation
    }

    # Set defaults for all display parameters
    set samplerate -1 ; set arrowmag -1 ; set scale -1
    if { ![string match {} $fixdisplayName] } {
	# Possibly keep current values of some display parameters
	global SubsampleRate FrameScale FrameArrowMag
	upvar $fixdisplayName fixdisplay
	if { [info exists fixdisplay(subsample)] \
		&& $fixdisplay(subsample)==1 } {
	    set samplerate $SubsampleRate
	}
	if { [info exists fixdisplay(arrowmag)] \
		&& $fixdisplay(arrowmag)==1 } {
	    set arrowmag $FrameArrowMag
	}
	if { [info exists fixdisplay(zoom)] \
		&& $fixdisplay(zoom)==1 } {
	    set scale $FrameScale
	}
    }

    # Use user filter (based on filename extension)?
    set tempfile [OMF_MmdispInputFilter $name]
    if { [string match {} $tempfile] } {
	# No filter match
	set workname $name
    } else {
	# Filter match
	set workname $tempfile
    }

    # Read file
    global DEBUGLVL
    if { 1 || $DEBUGLVL<10 } {
	ChangeMesh $workname $width $height $rot \
		$samplerate $arrowmag $scale
    } else {
	# For debugging; currently inactive
	set readtime [time {ChangeMesh $workname $width $height $rot \
		$samplerate $arrowmag $scale}]
	set readsecs [expr [lindex $readtime 0]/1000000.]
	set readsecs [format "%.2f" $readsecs]
	OMF_InfoDialog "File $name load time: $readsecs seconds" \
		"File load time"
    }

    # Delete tempfile, if any
    if { ![string match {} $tempfile] } {
	OMF_UnlinkFile $tempfile
    }

    incr FileId

    if {$ViewportHeight<1} {
	GetFrameSize ViewportWidth ViewportHeight
    }

    DrawFrame $canvas
    UpdateUserVariables
    PackViewport
}

proc SaveFileCallback { widget } {
    global ProgName
    global ViewportWidth ViewportHeight
    global MinViewportWidth MinViewportHeight
    global FrameSubsampleIncrement

    set errmsg {}

    # Get filename
    set filename [$widget GetFilename]

    # Store OMF type data & desc strings into omfspec array,
    # using the option name (w/o leading hyphen) as array index.
    foreach { name value } [$widget GetOptionState] {
	set omfspec([string range $name 1 end]) $value
    }
    foreach { name value } [$widget GetDescStrings] {
	set omfspec([string range $name 1 end]) $value
    }

    # Check to see if file already exits
    if { [file exists $filename] } {
	set answer [OMF_ModalDialog "File Save: File exists" {} \
		"File $filename already exists.\nOverwrite?" \
		{} {} 1 "Yes" "No" ]
	if { $answer != 0 } {
	    lappend errmsg "ERROR: File exists; no overwrite requested."
	    return $errmsg
	}
    }

    # Write file
    if { [catch {WriteMesh $filename $omfspec(datatype) $omfspec(gridtype) \
	    $omfspec(title) $omfspec(desc)} writeresult] } {
	lappend errmsg $writeresult
    }

    return $errmsg
}

proc OpenFileCallback { widget } {
    global ProgName
    global ViewportWidth ViewportHeight
    global MinViewportWidth MinViewportHeight
    global FrameSubsampleIncrement

    set errmsg {}

    # Get filename and "fix" data from File|Open widget
    set filename [$widget GetFilename]
    foreach { name value } [$widget GetOptionState] {
	# Strip leading hyphen from $name, and stuff
	# value into displayfix array
	set displayfix([string range $name 1 end]) $value
    }

    if { ![file exists $filename] } {
        OMF_InfoDialog "File $filename does not exist." \
		"$ProgName:Error" warning 10c
	lappend errmsg "ERROR: File does not exists"
    } elseif { ![file isfile $filename] } {
        OMF_InfoDialog "File $filename is not an ordinary file." \
		"$ProgName:Error" warning 10c
	lappend errmsg "ERROR: Special file"
    } elseif { ![file readable $filename] } {
        OMF_InfoDialog "Don't have read access to file $filename."
		"$ProgName:Error" warning 10c
	lappend errmsg "ERROR: No read access"
    } else {
        GetFrameRotation rot
        # If viewport has 0 area, fall back to default values (as needed).
        global DefaultViewportWidth DefaultViewportHeight
        if {$ViewportWidth<2} {
            set ViewportWidth $DefaultViewportWidth
        }
        if {$ViewportHeight<2} {
            set ViewportHeight $DefaultViewportHeight
        }

        # To be visually consistent, resize ViewportWidth/Height
        # to satisfy lower bounds.
        if {$ViewportWidth<$MinViewportWidth} { 
            set ViewportWidth $MinViewportWidth
        }
        if {$ViewportHeight<$MinViewportHeight} { 
            set ViewportHeight $MinViewportHeight
        }

	# Read file
	ReadFile $filename $ViewportWidth $ViewportHeight \
		$rot displayfix

        # Adjust subsample increment
        set newincr [GetMeshIncrement]
        if {$newincr>0} { set FrameSubsampleIncrement $newincr }
        # Initialize omfSavedDisplayState
        OMF_SaveDisplayState
    }
    return $errmsg
}

proc DialogCallback { widget actionid args } {
    # Re-enable menubutton when the dialog widget is destroyed
    if {[string match DELETE $actionid]} {
        eval [join $args]
	return
    }

    # Call appropriate action proc
    set errmsg "ERROR (proc DialogCallback): Invalid actionid: $actionid"
    if {[string match OPEN $actionid]} {
	set errmsg [OpenFileCallback $widget]
    } elseif {[string match SAVE $actionid]} {
	set errmsg [SaveFileCallback $widget]
    }

    return $errmsg
}

proc FileDialog { menuname itemlabel action } {
    if { [string match [$menuname entrycget $itemlabel -state] "disabled"] } {
	return   ;# Item not enabled
    }
    global ProgName FileId

    if {[string match OPEN $action]} {
	set dialogbox [omf_mmdopendlg new -callback DialogCallback \
		-dialog_title "$ProgName:Open File" \
		-filter "*.o?f *.svf *.vio" \
		-delete_callback_arg \
		"$menuname entryconfigure \{$itemlabel\} -state normal"]
    } elseif {[string match SAVE $action]} {
	set dialogbox [omf_mmdsavedlg new -callback DialogCallback \
		-dialog_title "$ProgName:Save File" \
		-filter "*.o?f *.svf *.vio" \
		-mesh_update_variable FileId \
		-mesh_update_proc OMF_MeshInfo \
		-delete_callback_arg \
		"$menuname entryconfigure \{$itemlabel\} -state normal"]
    } else {
	OMF_NonFatalError "Invalid action code: $action" \
		"Error in FileDialog proc"
	return
    }

    # Set icon
    MJD_SetIcon [$dialogbox cget -winpath]

    # Disable "File|Open" menubutton
    $menuname entryconfigure $itemlabel -state disabled
}

proc PrintDialogTrace {name elt op} {
    global omfPrint
    if {[string compare "OK" $omfPrint(done)]==0} {
	OMF_PushCursor  ;# Save current cursor & display "watch" cursor
	# Wrap following in a catch, to insure OMF_PopCursor gets called.
	if [catch {
	    OMF_SoftUpdate
	    OMF_PrintCanvas omfPrint
	} errmsg ] {
	    OMF_NonFatalError "Error in PrintDialogTrace: $errmsg"
	}
        OMF_PopCursor   ;# Replace original cursor
    }
}

proc PrintDisplay { menuname itemlabel } {
# Allow user to print display
    if { [string match [$menuname entrycget $itemlabel -state] "disabled"] } {
	return   ;# Item not enabled
    }
    global ProgName omfPrint

    # Update omfPrint variable
    set omfPrint(inname) [GetMeshName]

    # Open print dialog
    OMF_PrintDialog ".w${ProgName}PrintDisplay" \
            "$ProgName:Print Display" omfPrint
    trace variable omfPrint(done) w PrintDialogTrace
    # The following binding gets called for both the toplevel
    # level window and _all_ of its child windows!
    # Use quotes instead of braces to properly interpolate
    # $menuname and $itemlabel
    bind ".w${ProgName}PrintDisplay" <Destroy> "+
        # Ignore child window <Destroy> events
        set childwin \
          \"\[string compare \".w${ProgName}PrintDisplay\" \"%W\"\]\"
        if \$childwin then continue
        $menuname entryconfigure $itemlabel -state normal
        trace vdelete omfPrint(done) w PrintDialogTrace
    "
    $menuname entryconfigure $itemlabel -state disabled
}

###################### WIDGETS (Definitions & Bindings) #############
# Canvas widget
frame .canvas_frame -borderwidth 0
canvas .canvas_frame.canvas -background white -confine 1 \
        -borderwidth 0 -highlightthickness 0 \
        -xscrollcommand {.canvas_frame.xscroll set} \
        -yscrollcommand {.canvas_frame.yscroll set} \
        -width 0 -height 0

frame .canvas_frame.yscrollbox -borderwidth 0 ;# This needs to be created
                       # *before* yscroll so it gets mapped under yscroll.
scrollbar .canvas_frame.xscroll -troughcolor $omfTroughColor \
           -command ".canvas_frame.canvas xview"  -orient horizontal
scrollbar .canvas_frame.yscroll -troughcolor $omfTroughColor \
           -command ".canvas_frame.canvas yview"  -orient vertical
pack .canvas_frame.canvas -side left -fill both -expand 1
set ScrollCrossDim [expr \
        [.canvas_frame.yscroll cget -width] \
        + 2*[.canvas_frame.yscroll cget -borderwidth] \
        + 2*[.canvas_frame.yscroll cget -highlightthickness]]
frame .canvas_frame.yscrollbox.corner -borderwidth 2 -relief raised \
        -width $ScrollCrossDim -height $ScrollCrossDim
pack .canvas_frame.yscrollbox.corner -side bottom
# The yscrollbox.corner is used if both x & y scrollbars are displayed.
# By making the scrollbars symmetric, it eases the calculations needed
# to keep the same part of the image displayed under rotations.
# NOTE: The yscroll widget is directly packed into .canvas_frame if
#    xscroll is not displayed, but packed into .canvas_frame.yscrollbox
#    if xscroll is displayed.  See 'proc UpdateScrollBarDisplay'.

# NOTE: .canvas_frame.canvas <Configure> binding is set below in
#       INITIALIZATION section, and may get temporarily unset
#       and set in other places as well.

# Option frame and widgets
frame .ctrlbar -borderwidth 2 -relief raised
trace variable omfCtrlBarState w OMF_ChangeCtrlBarState
proc OMF_ChangeCtrlBarState { name elt op } {
    global omfCtrlBarState
    if {$omfCtrlBarState} {
        pack .ctrlbar -side top -after .menubar -fill x -expand 0
    } else {
        pack forget .ctrlbar
	focus .canvas_frame     ;# Shift focus off .ctrlbar
    }
    OMF_SetGeometryRequirements
}

# Rotation
frame .ctrlbar.rotation -borderwidth 2 -relief ridge
canvas .ctrlbar.rotation.coords -width 48 -height 0 -bg white
# The -width 48 is just a size guess; the canvas gets enlarged
# to a square region that just fills the height required
# by the other control bar lines. NOTE: The -height 0 value
# seems likely to be best behaved value wrt the <Configure>
# bindings.
pack .ctrlbar.rotation.coords -side left -fill y
proc UpdateCoordDisplay { newrot } {
    set win .ctrlbar.rotation.coords
    ShowCoords $win $newrot
}
trace variable DisplayRotation w DisplayRotationTrace
bind .ctrlbar.rotation.coords <Button-1> {
    set temp [expr (round($DisplayRotation/90.)+1)%%4]
    set DisplayRotation [expr $temp*90]
}
bind .ctrlbar.rotation.coords <Key-Left> {
    # Same as <Button-1>
    set temp [expr (round($DisplayRotation/90.)+1)%%4]
    set DisplayRotation [expr $temp*90]
}
bind .ctrlbar.rotation.coords <Button-3> {
    set temp [expr (round($DisplayRotation/90.)-1)%%4]
    set DisplayRotation [expr $temp*90]
}
bind .ctrlbar.rotation.coords <Key-Right> {
    # Same as <Button-3>
    set temp [expr (round($DisplayRotation/90.)-1)%%4]
    set DisplayRotation [expr $temp*90]
}
bind .ctrlbar.rotation.coords <Configure> {
    set bw [%W cget -borderwidth]
    set ht [%W cget -highlightthickness]
    set size [expr %h - 2*$bw - 2*$ht]
    set cwidth [%W cget -width]
    set cheight [%W cget -height]
    if {$cwidth!=$size || $cheight!=$size} {
        %W configure -width $size -height $size
    }
}

# Control bar row 1 frame; Holds "Subsample" and "Arrow mag" controls
frame .ctrlbar.row1

# Subsampling
OMF_MakeEntryScaleWindow .ctrlbar.subsample "Subsample:" 10 SubsampleRate 5 \
        50 0 10 $FrameSubsampleIncrement SetSubsamplingRate
trace variable FrameSubsampleIncrement w UpdateSubsampleIncrement
proc UpdateSubsampleIncrement { name elt op } {
    upvar $name newstep
    if {[info exists newstep]} {
        .ctrlbar.subsample.scale configure -resolution  $newstep
    }
}

# Arrow mag
OMF_MakeEntryWindow .ctrlbar.arrowmag " Arrow mag:" 0 FrameArrowMag \
	5 0 ridge 1
trace variable FrameArrowMag w SetArrowMagTrace

# Frame scale
OMF_MakeEntryScaleWindow .ctrlbar.framescale "Zoom:" 10 FrameScale 5 \
        50 0.5 100 .5 SetScale

### .ctrlbar packing
pack .ctrlbar.rotation   -side left -fill y
pack .ctrlbar.row1       -side top -fill x
pack .ctrlbar.subsample  -in .ctrlbar.row1 -side left  -fill both -expand 1 
pack .ctrlbar.arrowmag   -in .ctrlbar.row1 -side right -fill y

pack .ctrlbar.framescale -side top -fill x

########################### MENUS ###################################
frame .menubar -borderwidth 2 -relief raised
menubutton .file -text "File" -menu .file.menu -underline 0
menu .file.menu -tearoff 0
.file.menu add command -label "Open..." -underline 0 \
        -command { FileDialog .file.menu "Open..." OPEN }
## Note: If you change this menu item, then you must also change the
##       key binding with Ctrl-o; See the section "ADDITIONAL KEY BINDINGS"
.file.menu add command -label "Clear" -underline 0 \
        -command { ClearFrame }
.file.menu add command -label "Save as..." -underline 0 \
	-command { FileDialog .file.menu "Save as..." SAVE }
## Note: If you change this menu item, then you must also change the
##       key binding with Ctrl-s; See the section "ADDITIONAL KEY BINDINGS"
.file.menu add command -label "Print..." -underline 0 \
        -command { PrintDisplay .file.menu "Print..." }
## Note: If you change this menu item, then you must also change the
##       key binding with Ctrl-p; See the section "ADDITIONAL KEY BINDINGS"
.file.menu add separator
.file.menu add command -label "Exit" -command { exit } -underline 1
wm protocol . WM_DELETE_WINDOW { exit }

menubutton .view -text "View" -menu .view.menu -underline 0
menu .view.menu -tearoff 0
.view.menu add command -label "Wrap Display" -command { WrapDisplay } \
        -underline 0
.view.menu add command -label "Fill Display" -command { FillDisplay } \
        -underline 0
.view.menu add command -label "Rotate ccw" -underline 0 \
        -command {set DisplayRotation [expr round($DisplayRotation+90)%360]}
.view.menu add command -label "Rotate cw" -underline 7 \
        -command {set DisplayRotation [expr round($DisplayRotation-90)%360]}

menubutton .options -text "Options" -menu .options.menu -underline 0
menu .options.menu -tearoff 0
.options.menu add check -label "Control Bar" -underline 0 \
        -variable omfCtrlBarState

menubutton .help -text "Help" -menu .help.menu -underline 0
menu .help.menu -tearoff 0
.help.menu add command -label "About" -underline 0 \
        -command { AboutInfo .help.menu "About" }
.help.menu add command -label "Credits" -underline 0 \
        -command { CreditsInfo .help.menu "Credits" }
.help.menu add command -label "License" -underline 0 \
        -command { LicenseInfo .help.menu "License" }
.help.menu add command -label "On Mmdisp..." -underline 0 \
	-command {OMFHB_Browser [OMFURL_FileToUrl [file join $omfmmd_library \
                  [OMFMMD_FileNameBase][OMFMMD_Version] help]]}

### .menubar packing
pack .file .view .options -side left -in .menubar
pack .help -side right -in .menubar

####################### ADDITIONAL KEY BINDINGS #####################
# Allow the canvas frame to acquire keyboard focus, and prohibit
# canvas or canvas scrollbars from same (all necessary key bindings
# are bound to the canvas frame.
set cf .canvas_frame
$cf.xscroll configure -takefocus 0
$cf.yscroll configure -takefocus 0
$cf.canvas  configure -takefocus 0
$cf configure -takefocus 1

proc Zoom { amt } { global FrameScale; SetScale [expr $amt*$FrameScale] }
bind $cf <Key-Escape>      { OMF_RecoverDisplayState }
bind $cf <Key-Prior>       { Zoom $SmallZoom }
bind $cf <Shift-Key-Prior> { Zoom $LargeZoom }
bind $cf <Key-Next>        { Zoom [expr 1./$SmallZoom] }
bind $cf <Shift-Key-Next>  { Zoom [expr 1./$LargeZoom] }
## NOTE: SunOS apparently doesn't have Page_Up & Page_Down.  For the
## record, Prior == Page_Up, and Next == Page_Down.
# Ditto for KeyPad; Note that the <Shift> keypad values are funky,
# and might be wrong for some keyboards.
catch {bind $cf <Key-KP_Prior>    { Zoom $SmallZoom }}
catch {bind $cf <Shift-Key-KP_9>  { Zoom $LargeZoom }}
catch {bind $cf <Key-KP_Next>     { Zoom [expr 1./$SmallZoom] }}
catch {bind $cf <Shift-Key-KP_3>  { Zoom [expr 1./$LargeZoom] }}

bind $cf <Control-Key-f> { FillDisplay }
bind $cf <Control-Key-w> { WrapDisplay }
bind $cf <Key-Home>      { FillDisplay ; WrapDisplay }
catch {bind $cf <Key-KP_Home>   { FillDisplay ; WrapDisplay }}
bind $cf <Control-Key-o> { FileDialog .file.menu "Open..." OPEN }
bind $cf <Control-Key-s> { FileDialog .file.menu "Save as..." SAVE }
bind $cf <Control-Key-p> { PrintDisplay .file.menu "Print..." }
## Note: The last two depend upon the text in the pull down menus!

bind $cf <Control-Key-r> { ;# Ctrl-r
    set DisplayRotation [expr round($DisplayRotation+90)%%360]
}
bind $cf <Control-Key-R> { ;# Shift-Ctrl-R
    set DisplayRotation [expr round($DisplayRotation-90)%%360]
}
bind $cf <Key-Insert> { 
    set newrate [expr $SubsampleRate-1]
    if {$newrate<0} {set newrate 0}
    SetSubsamplingRate $newrate
}
bind $cf <Key-Delete> {SetSubsamplingRate [expr $SubsampleRate+1]}
bind $cf <Shift-Key-Insert> {SetSubsamplingRate [expr $SubsampleRate/2.]}
bind $cf <Shift-Key-Delete> {SetSubsamplingRate [expr $SubsampleRate*2]}
# Ditto for KeyPad; Note that the <Shift> keypad values are funky,
# and might be wrong for some keyboards.
catch { bind $cf <Key-KP_Insert> { 
    set newrate [expr $SubsampleRate-1]
    if {$newrate<0} {set newrate 0}
    SetSubsamplingRate $newrate
}   }
catch {bind $cf <Key-KP_Delete> {SetSubsamplingRate [expr $SubsampleRate+1]}}
catch {bind $cf <Shift-Key-KP_0> {
    SetSubsamplingRate [expr $SubsampleRate/2.]
}   }
catch {bind $cf <Shift-Key-KP_Decimal> {
    SetSubsamplingRate [expr $SubsampleRate*2]
}   }

# Proc to scroll canvas from keyboard.  direction should be one of
# left|right|up|down, and stride should be one of units|pages.
proc KeyboardScrollCanvas { direction stride } {
    global canvas
    switch -- $direction {
        left     { $canvas xview scroll -1 $stride }
        right    { $canvas xview scroll  1 $stride }
        up       { $canvas yview scroll -1 $stride }
        down     { $canvas yview scroll  1 $stride } 
        default  {
            OMF_FatalError "Invalid direction ($direction) detected in \
                    proc KeyboardScrollCanvas"
        }
    }
}
bind $cf <Key-Left>  { KeyboardScrollCanvas left  units }
bind $cf <Key-Right> { KeyboardScrollCanvas right units }
bind $cf <Key-Up>    { KeyboardScrollCanvas up    units }
bind $cf <Key-Down>  { KeyboardScrollCanvas down  units }
bind $cf <Shift-Key-Left>  { KeyboardScrollCanvas left  pages }
bind $cf <Shift-Key-Right> { KeyboardScrollCanvas right pages }
bind $cf <Shift-Key-Up>    { KeyboardScrollCanvas up    pages }
bind $cf <Shift-Key-Down>  { KeyboardScrollCanvas down  pages }
# Ditto for KeyPad; Note that the <Shift> keypad values are funky,
# and might be wrong for some keyboards.
catch {bind $cf <Key-KP_Left>  { KeyboardScrollCanvas left  units }}
catch {bind $cf <Key-KP_Right> { KeyboardScrollCanvas right units }}
catch {bind $cf <Key-KP_Up>    { KeyboardScrollCanvas up    units }}
catch {bind $cf <Key-KP_Down>  { KeyboardScrollCanvas down  units }}
catch {bind $cf <Shift-Key-KP_4> { KeyboardScrollCanvas left  pages }}
catch {bind $cf <Shift-Key-KP_6> { KeyboardScrollCanvas right pages }}
catch {bind $cf <Shift-Key-KP_8> { KeyboardScrollCanvas up    pages }}
catch {bind $cf <Shift-Key-KP_2> { KeyboardScrollCanvas down  pages }}

# Try to handle <Shift-Key-Tab> variants.  The [bind all <Shift-Tab>]
# parameter expands to the proc that is the current binding to
# <Shift-Tab> (probably 'tkTabToWindow [tk_focusPrev %W]').  Some
# X servers return a distinct keysym for the Shift-Tab event, e.g.,
# ISO_Left_Tab _instead_ Tab.  One can use the X utility program
# xev to see what keysyms are being generated.
#  One problem that occurs, is if one is running a binary on one
# machine, say, a Sun, that does not know the ISO_Left_Tab keysym
# (cf. the system <X11/keysymdef.h> file), and displaying on a
# different system, say recent Linux machine, that does.  Then the
# X server on the Linux machine generates the keysym number corresponding
# to ISO_LEFT_TAB (0xfe20), but the program built on the Sun doesn't
# know the symbolic interpretation for that keysym.  Therefore we
# provide a numeric binding if the symbolic binding is not known.  (I'm
# assuming the symbolic bindings, if defined, are the same across all
# platforms.)
#  In addition to xev, the following wish script is useful in diagnosing
# and solving such problems:
#
# bind . <KeyPress> {puts stdout {%%K=%K, %%A=%A, %%k=%k, %%s=%s, %%N=%N.}}
#
# This prints to stdout the  keysym (%K), ASCII character (%A), keycode (%k)
# state (%s, which for key events is modifier information, like shift key
# status), and keysym as a decimal number (%N).  The last should be defined
# even when the first isn't.          -mjd, Sep-1997
#
if {[catch {bind all <Shift-Key-ISO_Left_Tab> [bind all <Shift-Tab>]}]} {
    # ISO_Left_Tab probably not defined.  Try numeric binding
    catch {
	bind all <Shift-Any-Key> \
		"+if {\"%N\"==65056} \{[bind all <Shift-Tab>]\}"
	# ISO_Left_Tab = 0xFE20 = 65056
	# Man, talk about quoting hell.  We need the braces on the
	# second term of the "if" to keep "if" happy, but we need
	# to get the [bind all <Shift-Tab>] expanded at bind time to
	# get the imbedded %W (from inside [bind all <Shift-Tab]) in
	# place for binding substitution...
    }
}
if {[catch {bind all <Shift-Key-3270_BackTab> [bind all <Shift-Tab>]}]} {
    # 3270_BackTab probably not defined.  Try numeric binding
    catch {
	bind all <Shift-Any-Key> \
		"+if {\"%N\"==64773} \{[bind all <Shift-Tab>]\}"
	# 3270_BackTab = 0xFD05 = 64773
	# See the notes in the previous stanza about "quoting hell."
    }
}

###################### ADDITIONAL MOUSE BINDINGS ####################
### Canvas rectangle zoom-in
set omfZoomBoxForeColorA #FF0000     ;# Zoom box foreground line color,
set omfZoomBoxForeColorB #0000CD     ;# A -> enlarge, B -> contract
# The first is "red", the second "MediumBlue".  For reasons unknown
# to me (mjd 97/2/22), when drawing on an 8-bit pseudocolor visual
# with a shared colormap, under interactive conditions (redrawing
# zoom boxes with mouse motion) it took noticeable longer to erase
# and draw "red" boxes than "MediumBlue" boxes.  I discovered that the
# difference goes away if "red" is changed to, e.g., "IndianRed", or to
# the form #xxx.  I also discovered that you can apparently find out
# what color you will actually get with the command "winfo rgb . color",
# where color can be in any of the usual Tk color formats.  For example,
#   %winfo rgb . MediumBlue
#  > 0 0 52685
#   %winfo rgb . #FF0000
#  > 65535 0 0
#   %winfo rgb . #FE0000
#  > 65535 0 0
#
set omfZoomBoxForeThick  1       ;# Zoom box foreground line thickness
set omfZoomBoxBackColor  white   ;# Zoom box background line color
set omfZoomBoxBackThick  3       ;# Zoom box background line thickness
set omfZoomBoxMinWidth   5;# This must be >0.
set omfZoomBoxArrowSize  12
set omfZoomDirection     0       ;# When active, should be +1 => enlarge
                                  #                     or -1 => contract
set omfZoomBoxLoiterTime 3000    ;# Number of milliseconds to keep zoom box
                                  # visible after zooming.
proc OMF_InitZoomBox { v1x v1y direction } {
    global canvas omfP1x omfP1y omfZoomDirection
    if {$omfZoomDirection!=0} return  ;# Zoom (other direction) already
    set omfZoomDirection $direction   ;# in progress
    set omfP1x [$canvas canvasx $v1x]
    set omfP1y [$canvas canvasy $v1y]
}

proc OMF_StrokeZoomBox { x1 y1 x2 y2 boxtag} {
    global omfZoomBoxBackColor omfZoomBoxBackThick
    global omfZoomBoxForeThick omfZoomDirection
    global omfZoomBoxForeColorA omfZoomBoxForeColorB
    global canvas
    if {$omfZoomDirection==0} return
    if {$omfZoomDirection>0} { set forecolor $omfZoomBoxForeColorA } \
            else             { set forecolor $omfZoomBoxForeColorB }
    $canvas create rectangle $x1 $y1 $x2 $y2 \
            -outline $omfZoomBoxBackColor -width $omfZoomBoxBackThick \
            -tag $boxtag
    $canvas create rectangle $x1 $y1 $x2 $y2 \
            -outline $forecolor -width $omfZoomBoxForeThick \
            -tag $boxtag
}

proc OMF_DrawZoomBox { v2x v2y } {
    global omfP1x omfP1y canvas
    global omfZoomBoxMinWidth
    global omfZoomDirection omfZoomBoxForeColorA omfZoomBoxForeColorB
    if {$omfZoomDirection==0} return
    if {$omfZoomDirection>0} {
        set forecolor $omfZoomBoxForeColorA
    } else {
        set forecolor $omfZoomBoxForeColorB
    }
    global omfZoomBoxBackColor omfZoomBoxForeThick omfZoomBoxBackThick
    global omfZoomBoxArrowSize
    $canvas delete zoombox ;# erase old zoom box
    set p2x [$canvas canvasx $v2x]
    set p2y [$canvas canvasy $v2y]
    set zw [expr abs($p2x-$omfP1x)]
    set zh [expr abs($p2y-$omfP1y)]
    if {$zw<$omfZoomBoxMinWidth || $zh<$omfZoomBoxMinWidth} {
        global FrameScale FrameScaleScale FrameScaleEntry
	set FrameScale [OMF_FormatNumber $FrameScale]
        if {$FrameScaleEntry!=$FrameScale} {
            set FrameScaleScale $FrameScale ;# FSEntry should be set
            set FrameScaleEntry $FrameScale ;# *after* FSScale.
        }
        return
    }
    OMF_StrokeZoomBox $omfP1x $omfP1y $p2x $p2y zoombox

    # Determine which side (x- or y-) will limit expansion
    set vw [$canvas cget -width]
    set vh [$canvas cget -height]
    if {$omfZoomDirection>0} {
        # Make adjustments to viewport dimensions under the
        # assumption that any non-active scroll bars will be
        # turn on as a result of the zoom.
        global XScrollState YScrollState ScrollCrossDim
        if {!$XScrollState} {
            set vh [expr $vh-$ScrollCrossDim]
        }
        if {!$YScrollState} {
            set vw [expr $vw-$ScrollCrossDim]
        }
    }
    if {$vw<1} {set vw 1} ; if {$vh<1} {set vh 1} ;# Safety
    set xlimit [expr $vw*$zh]; set xfudge [expr 2*$vw]
    set ylimit [expr $vh*$zw]; set yfudge [expr 2*$vh]
    if {$omfZoomDirection<0} {
        set temp $xlimit; set xlimit $ylimit; set ylimit $temp;
        set temp $xfudge; set xfudge $yfudge; set yfudge $temp;
    }

    # Draw zoom direction arrows on expansion limit sides
    if {$omfP1x<$p2x} { set xmin $omfP1x; set xmax $p2x    } \
            else      { set xmin $p2x;    set xmax $omfP1x }
    if {$omfP1y<$p2y} { set ymin $omfP1y; set ymax $p2y    } \
            else      { set ymin $p2y;    set ymax $omfP1y }
    set xmid [expr round(($xmin+$xmax)/2.)]
    set ymid [expr round(($ymin+$ymax)/2.)]
    set arrlength [expr round($omfZoomBoxArrowSize)]
    if {$omfZoomDirection>=0} { set offset $arrlength } \
            else              { set offset -$arrlength }
    set arrheadsize [expr round($arrlength/2)]
    set arrheadshape "$arrheadsize  $arrheadsize [expr $arrheadsize/2]"
    if { $ylimit < $xlimit + $xfudge} {
        # Draw arrows on top & bottom sides
        $canvas create line \
                $xmid [expr $ymin+$offset] $xmid $ymin \
                -fill $forecolor -width $omfZoomBoxForeThick \
                -arrow last -arrowshape $arrheadshape -tag zoombox
        $canvas create line \
                $xmid [expr $ymax-$offset] $xmid $ymax \
                -fill $forecolor -width $omfZoomBoxForeThick \
                -arrow last -arrowshape $arrheadshape -tag zoombox
    }
    if { $xlimit < $ylimit + $yfudge} {
        # Draw arrows on left & right sides
        $canvas create line \
                [expr $xmin+$offset] $ymid $xmin $ymid \
                -fill $forecolor -width $omfZoomBoxForeThick \
                -arrow last -arrowshape $arrheadshape -tag zoombox
        $canvas create line \
                [expr $xmax-$offset] $ymid $xmax $ymid \
                -fill $forecolor -width $omfZoomBoxForeThick \
                -arrow last -arrowshape $arrheadshape -tag zoombox
    }
    # Reflect anticipated zoom state in scale widgets
    if {$vw>0 && $vh>0 && $zw>0 && $zh>0} { ;# Safety check
        if {$ylimit < $xlimit} {
            set scale [expr double($vh)/double($zh)]
        } else {
            set scale [expr double($vw)/double($zw)]
        }
        if {$omfZoomDirection<0} { set scale [expr 1./double($scale)] }
        global FrameScale FrameScaleScale FrameScaleEntry
        global omfStdNumericFormat
	set newsize [OMF_FormatNumber [expr $FrameScale*$scale]]
        set FrameScaleScale $newsize  ;# FSEntry should be set *after*
        set FrameScaleEntry $newsize  ;# FSScale.
    }
}
proc OMF_DoZoom { v2x v2y } {
    global canvas omfP1x omfP1y
    global FrameScale
    global omfZoomBoxMinWidth omfZoomDirection
    if {$omfZoomDirection==0} return
    $canvas delete zoombox ;# erase old zoom box
    set p2x [$canvas canvasx $v2x]
    set p2y [$canvas canvasy $v2y]
    set zwidth  [expr abs($p2x-$omfP1x)]
    set zheight [expr abs($p2y-$omfP1y)]
    if {$zwidth<$omfZoomBoxMinWidth \
            || $zheight<$omfZoomBoxMinWidth} {
        unset omfP1x omfP1y
        set omfZoomDirection 0
        return
    }
    set vwidth  [$canvas cget -width]
    set vheight [$canvas cget -height]
    set x_scrollstate_adj 0
    set y_scrollstate_adj 0
    if {$omfZoomDirection>0} {
        # Make adjustments to viewport dimensions under the
        # assumption that any non-active scroll bars will be
        # turn on as a result of the zoom.
        global XScrollState YScrollState ScrollCrossDim
        if {!$XScrollState} {
            set vheight [expr $vheight-$ScrollCrossDim]
            set y_scrollstate_adj [expr -$ScrollCrossDim]
        }
        if {!$YScrollState} {
            set vwidth [expr $vwidth-$ScrollCrossDim]
            set x_scrollstate_adj [expr -$ScrollCrossDim]
        }
    }
    set xscale [expr double($vwidth)/double($zwidth)]
    set yscale [expr double($vheight)/double($zheight)]
    if {$xscale<0.001 || $xscale>1000  || $yscale<0.001 || $yscale>1000} {
        # Safety check
        unset omfP1x omfP1y
        set omfZoomDirection 0
        return
    }
    if {$omfZoomDirection<0} {
        set xscale [expr 1./double($xscale)]
        set yscale [expr 1./double($yscale)]
    }
    if {$xscale<$yscale} {
        set scale $xscale
        set xadj 0
        set yadj [expr round(($vheight/$scale-$zheight)/2.)]
        # xadj & yadj are 1/2 the extra space left in each
        # dimension after scaling by $scale.  This is used
        # to center the zoomed region after scaling.
    } else {
        set scale $yscale
        set xadj [expr round(($vwidth/$scale-$zwidth)/2.)]
        set yadj 0
    }

    # Move and zoom
    if {$omfP1x<$p2x} {set xmin $omfP1x} else {set xmin $p2x}
    if {$omfP1y<$p2y} {set ymin $omfP1y} else {set ymin $p2y}
    GetFrameSize cwidth cheight
    if {$cwidth<1 || $cheight<1} {
        unset omfP1x omfP1y
        set omfZoomDirection 0
        return
    }
    if {$omfZoomDirection>=0} {
        set xadj [expr round(($vwidth/$scale-$zwidth)/2.)]
        set yadj [expr round(($vheight/$scale-$zheight)/2.)]
        set xbase [expr $xmin-$xadj]  ;# Canvas coordinates of new
        set ybase [expr $ymin-$yadj]  ;# upper left hand corner
    } else {
        set xadj [expr round(($vwidth-$zwidth/$scale)/2.)]
        set yadj [expr round(($vheight-$zheight/$scale)/2.)]
        set xbase [$canvas canvasx 0]
        set xbase [expr $xbase+$xadj-double($xmin-$xbase)/double($scale)]
        set ybase [$canvas canvasy 0]
        set ybase [expr $ybase+$yadj-double($ymin-$ybase)/double($scale)]
    }

    
    OMF_SaveDisplayState
    SetFrameScale [expr $scale*$FrameScale]
    GetFrameScale temp  ;  set FrameScale [OMF_FormatNumber $temp]

    DrawFrame $canvas
    PackViewport "" "" 0
    $canvas xview moveto [expr double($xbase)/double($cwidth)]
    $canvas yview moveto [expr double($ybase)/double($cheight)]
    #UpdateWidgetGeometries  ;# This doesn't seem to actually be needed...
    #update idletasks ;# ditto

    # Draw zoom "loiter" box
    set a1x [expr $omfP1x*$scale]
    set a1y [expr $omfP1y*$scale]
    set a2x [expr $p2x*$scale]
    set a2y [expr $p2y*$scale]
    if {$omfZoomDirection>0} {
        # Adjust coordinates so box fits inside viewport
        if {$a2x<$a1x} {set temp $a1x; set a1x $a2x; set a2x $temp}
        if {$a2y<$a1y} {set temp $a1y; set a1y $a2y; set a2y $temp}
        set a1x [expr floor($a1x)] ; set a1y [expr floor($a1y)]
        set a2x [expr  ceil($a2x)] ; set a2y [expr  ceil($a2y)]
        set vwidth  [$canvas cget -width]
        set vheight [$canvas cget -height]
        set xmin [$canvas canvasx 0]
        set ymin [$canvas canvasy 0]
        set xmax [expr [$canvas canvasx $vwidth]-1]
        set ymax [expr [$canvas canvasy $vheight]-1]
        if {$a1x<$xmin} {set a1x $xmin}
        if {$a1y<$ymin} {set a1y $ymin}
        if {$a2x>$xmax} {set a2x $xmax}
        if {$a2y>$ymax} {set a2y $ymax}
    }
    global UpdateViewCount
    set boxtag "afterzoombox$UpdateViewCount"
    # The above creates an essentially unique tagname for the zoom
    # loiter box.  If all loiter boxes shared the same tag, then if
    # the user did a second zoom too quick, then the second zoom loiter
    # box could be deleted by the command deleting the first.  OK, this
    # is not a serious problem, but its easy to fix, and I'm here now
    # anyway...
    OMF_StrokeZoomBox $a1x $a1y $a2x $a2y $boxtag
    unset omfP1x omfP1y
    set omfZoomDirection 0
    global omfZoomBoxLoiterTime
    after $omfZoomBoxLoiterTime "$canvas delete $boxtag"
}
proc OMF_CenterView { vx vy } {
    global canvas
    set px [$canvas canvasx $vx]
    set py [$canvas canvasy $vy]
    GetFrameSize cwidth cheight
    if {$cwidth<1 || $cheight<1} return
    set xcenter [expr $px/$cwidth]
    if {$xcenter<0} {set xcenter 0}
    if {$xcenter>1} {set xcenter 1}
    set ycenter [expr $py/$cheight]
    if {$ycenter<0} {set ycenter 0}
    if {$ycenter>1} {set ycenter 1}
    SetScrollCenter $xcenter $ycenter
}
bind $canvas <Button-1> { OMF_InitZoomBox %x %y 1 }
bind $canvas <Button-3> { OMF_InitZoomBox %x %y -1 }
bind $canvas <Button1-Motion> { OMF_DrawZoomBox %x %y }
bind $canvas <Button3-Motion> { OMF_DrawZoomBox %x %y }
bind $canvas <ButtonRelease> { OMF_DoZoom %x %y }
bind $canvas <ButtonRelease-2> { OMF_CenterView %x %y}
bind $canvas <Shift-ButtonRelease-1> { OMF_CenterView %x %y}
# Hmm, can't seem to find a home for a CenterView binding

########################### TOP_LEVEL PACKING #######################
pack .menubar -side top -fill x -expand 0
pack .canvas_frame -side top -fill both -expand 1
# Note: Any extra vertical spacing is divvied up equally among
#   all the first level widgets that have the -expand 1 option.
#   The net result is that frame-canvas resizing won't work properly
#   if any widget besides .canvas_frame has -expand 1

# Ctrlbar
if {$omfCtrlBarState} {
    pack .ctrlbar -side top -after .menubar -fill x -expand 0
}

########################### INITIALIZATION ##########################
# Determine minimum viewport size, and get fixed widget geometry values
OMF_SetGeometryRequirements

# Initialize global variables
set DefaultViewportWidth  [expr $DefaultWindowWidth-10] 
set DefaultViewportHeight [expr $DefaultWindowHeight-$RootHeadHeight-20]
# The '-10' & '-20' in the above lines is just so we don't fill up the
# *whole* display screen (in case the display *is* only 640x480).

# Disable $canvas <Configure> binding while we set things up
bind $canvas <Configure> ""

# Extract working directory and filename from command line argument
set firstpath {}
set firstfile [GetFirstFileName]
if [file isdirectory $firstfile] {
    # No file, just directory
    set firstpath $firstfile
    set firstfile {}
} else {
    # File and potentially directory
    set firstpath [file dirname $firstfile]
    set firstfile [file tail $firstfile]
}

# Change to requested working directory
cd $firstpath

# Load first file
ReadFile $firstfile

# Initialize and show coords display
set coords_win .ctrlbar.rotation.coords
WaitVisible $coords_win
global ColorMap DisplayRotatin
InitCoords $coords_win $ColorMap(base) $ColorMap(in) \
	$ColorMap(out) $DisplayRotation
UpdateCoordDisplay $DisplayRotation

# Release firstpath & firstfile variables
unset firstpath firstfile

# Adjust tabbing order
raise .ctrlbar
raise .ctrlbar.rotation.coords
raise .ctrlbar.subsample.value
raise .ctrlbar.subsample.scale
raise .ctrlbar.framescale.value
raise .ctrlbar.framescale.scale
raise .canvas_frame

# Bind canvas resize proc
update
bind $canvas <Configure> {
    set border [expr 2*[%W cget -borderwidth]+2*[%W cget -highlightthickness]]
    ResizeViewport [expr %w-$border] [expr %h-$border]
    # Note: The <Configure> resize request *includes* both -borderwidth
    #       & -highlightthickness!
}
focus .canvas_frame

# Initialize omfSavedDisplayState
OMF_SaveDisplayState

# Setup icon
MJD_SetIcon .

