# FILE: vecfile.tcl
#
# Tcl/Tk routines for vector file input
#

proc OMF_SvfIsType { filename chanid } {
    if { [gets $chanid line] <= 0 } { return 0; }
    # See if first line has "# SVF-2" -like string
    if { [regexp -nocase -- "^#\[ \t\]*svf-\[0-9\]+" $line] } { return 1; }

    # Otherwise, check file extension and first non-comment line
    if { ![string match {*.svf*} [string tolower $filename]] } { return 0; }
    while { [regexp -- "^\[ \t\]*#|^\[ \t\]*$" $line] } {
	# $line is a blank or comment line, so get another line
	if { [gets $chanid line] <= 0 } { return 0; }
    }
    # $line is now first "real" line
    set nf {([-+]?[.0-9]+[eEdD]?[-+]?[0-9]*)}
    ## Numeric field regexp pattern with grouping
    set ws "\[ \t\]"         ;# Whitespace regexp pattern
    set cmt {(#.*)?}         ;# Comment regexp pattern
    set datapat "^$ws*$nf$ws+$nf$ws+$nf$ws+$nf$ws+$nf\($ws+$nf\)?$ws*$cmt$"
    if { [regexp -- $datapat $line] } {
	return 1      ;# Line has proper data format
    }
    return 0
}

# Utility proc for OMF_SvfRead to process svf header lines
# Returns >0 if line contains auxilary info,
#          0 if line is uninterpreted valid (an ordinary comment),
#    or   -1 if line has bad format (not a comment line)
proc _omfProcessCommentLine { sfi_index line } {
    set ws "\[ \t\]"           ;# Whitespace regexp pattern
    set cmt {(#.*)?}           ;# Comment regexp pattern
    set blankpat "^$ws*$cmt$"  ;# Blank line
    set returncode "0"
    # If not in valid comment format, mark as bad.
    if { ![regexp -- $blankpat $line] } {
	set returncode "-1"
    } else {
	# Valid comment line.  Check for auxiliary information.
	# Boundary info
	if { [regexp -nocase -- "##$ws*boundary.*xy\[:\]?(.*)" \
		$line junk points] } {
	    # $points should be list of boundary points
	    set points [string trim $points]
	    OMF_SvfFileInputTA setbdry $sfi_index $points
	    incr returncode
	}
	# Grid step info
	if { [regexp -nocase -- "##$ws*grid$ws*step\[:\]?(.*)" \
		$line junk steps] } {
	    # $steps should be list of x, y, z steps
	    set steps [string trim $steps]
	    OMF_SvfFileInputTA stephints $sfi_index $steps
	    incr returncode
	}
	# Title info
	if { [regexp -nocase -- "##$ws*title$ws*\[:\]?(.*)" \
		$line junk title] } {
	    set title [string trim $title]
	    OMF_SvfFileInputTA settitle $sfi_index $title
	    incr returncode
	}
    }
    return $returncode
}

# Procedure to process an SVF file.  Returns the number of points, the
# number of lines, and the number of lines that couldn't be parsed,
# i.e., the third field should be 0 for a properly formatted file.
proc OMF_SvfRead { chanid sfi_index } {
    set commentcount 0 ; set badlinecount 0
    while { [gets $chanid line] >= 0 } {
	set linecode [OMF_SvfFileInputTA processline $sfi_index $line]
	if { [string match "1" $linecode] } { continue }
	## The above case is the most common, data line case.  The
	## parsing is done in C for speed.

	# Below follow the comment line or "bad" data line cases.
	if { [string match "0" $linecode] } {
	    _omfProcessCommentLine $sfi_index $line  ;#Comment line
	    incr commentcount
	} elseif { [string match "-1" $linecode] } {
	    incr badlinecount                        ;# Bad data line
	} else {
	    OMF_FatalError \
	    "Unknown return code from \"OMF_SvfFileInputTA processline\"" \
	    "Error in proc OMF_SvfRead" 
	}
    }
    set pointcount [OMF_SvfFileInputTA getpointcount $sfi_index]
    set linecount  [expr $pointcount + $commentcount + $badlinecount]
    return "$pointcount $linecount $badlinecount"
}


# BELOW is the original OMF_SvfRead, that parse the input file entirely
# in Tcl.  It is rather slow on older machines and big files, and so
# has been replaced by the routine above.
proc _OMF_SvfRead_Original { chanid sfi_index } {
    set linecount 0 ; set pointcount 0 ; set badlinecount 0
    set nf {([-+]?[.0-9]+[eEdD]?[-+]?[0-9]*)}
    ## Numeric field regexp pattern with grouping
    set ws "\[ \t\]"         ;# Whitespace regexp pattern
    set cmt {(#.*)?}         ;# Comment regexp pattern
    set datapat "^$ws*$nf$ws+$nf$ws+$nf$ws+$nf$ws+$nf\($ws+$nf\)?$ws*$cmt$"
    set blankpat "^$ws*$cmt$"
    while { [gets $chanid line] >= 0 } {
	incr linecount
	##_OMF_OriginalPuts stdout "Input line ->$line<-"

	# Extract numeric fields.  These should be in the order
	# "x y z mx my mz", but some older files (version Svf-01)
	# may have only "x y mx my mz".
	if { ![regexp -- $datapat $line junk x y z mx my junk2 mz ] } {
	    # No match.  If not a blank line, then mark as bad
	    if { ![regexp -- $blankpat $line] } {
		incr badlinecount
	    } else {
		# Blank line.  Check for auxiliary information
		# Boundary info
		if { [regexp -nocase -- "##$ws*boundary.*xy\[:\]?(.*)" \
			$line junk points] } {
		    # $points should be list of boundary points
		    set points [string trim $points]
		    OMF_SvfFileInputTA setbdry $sfi_index $points
		}
		# Grid step info
		if { [regexp -nocase -- "##$ws*grid$ws*step\[:\]?(.*)" \
			$line junk steps] } {
		    # $steps should be list of x, y, z steps
		    set steps [string trim $steps]
		    OMF_SvfFileInputTA stephints $sfi_index $steps
		}
	    }
	    continue;    ;# Read next line
	}
	# Good line;  Now, was this a 5-field or a 6-field match?
	if {[string match {} $mz]} {
	    # 5-field match
	    set mz $my; set my $mx; set mx $z
	    set z 0
	}

	# Add point to mesh
	incr pointcount
	OMF_SvfFileInputTA addpoint $sfi_index $x $y $z $mx $my $mz
	## _OMF_OriginalPuts stdout "Pos=($x,$y,$z) m=($mx,$my,$mz)"
    }
    ## OMF_NonFatalError \
    ##	    "READ RETURN: $pointcount $linecount $badlinecount" \
    ##	    "OMF_SvfRead Debub Output"
    return "$pointcount $linecount $badlinecount"
}
