/* FILE: grid.cc           -*-Mode: c++-*-
 *
 * This file contains class definitions for micromagnetic grid classes.
 *
 */

/* Revision history:
 *  7-Apr-96: Removed last vestiges of g2d file format.
 *  9-Jul-96: Added support for maximum angle difference determination.
 *  8-Jul-96: Revamped ODE solver.  Introduced 4th order Runge-Kutta
 *            solver with adaptive stepsize.
 * 30-Jun-96: Added Array2D<MagElt> support.
 */

#include <string.h>

#include "demag.h"
#include "misc.h"
#include "grid.h"
#include "magelt.h"
#include "vecio.h"
#include "vector.h"
#include "maganis.h"

/* End includes */

// Define UNIT_MAG to get speed-ups from assuming all magnetizations
// are unit vectors
#undef UNIT_MAG
#undef WEAK_BOUNDARY // 0.5  // Set boundary element reduced moments to this
                           // #undef or set to 0.0 to get sharp boundaries.
#undef EDGE_ADJUST // 1.0 // Alternate form of WEAK_BOUNDARY; Affects demag
                    // only, and only works with RECTANGULAR sample shapes!!!
#undef ROUGH_BOUNDARY
    // If #defined, then each boundary moment is independently multiplied
    // by an amount drawn from a uniform distribution on [0,1].

// ODE parameters
const double Grid2D::GyRatio=2.21e5; // Gyromagnetic ratio, m/A.s
const double Grid2D::DampCoef=0.5;   // Damping coefficient, dimensonless
// The damping coefficient can be obtained via FMR studies as
//  DampCoef=(Delta H).GyRatio/(2.(fmr_driving_frequency)).
//  'Delta H' here is the FMR line width.  In the literature,
//  GyRatio is denoted by \gamma, fmr_driving_frequency is \omega_0,
//  and DampCoef is often written \alpha or \alpha/(1+\alpha).
// Realistic values for DampCoef are in the range [.004,.15], but for
//  fast convergence of the ODE, DampCoef=0.5 seems close to optimal.
// dm/dt = (-GyRatio.Ms)(mxh)-(DampCoef.GyRatio.Ms)(mx(mxh))
//  This program actually solves RHS=(-1/DampCoef)(mxh)-mx(mxh).
//  The routine Grid2D::GetTimeStep() does the conversion from StepSize
//  to actual time step in seconds, via
//              time_step = StepSize/(DampCoef.GyRatio.Ms)
//  See also PRECESSION_RATIO below.

// #define exactly one of the following to select the coefficient
// to use with the internal demag routine.
#undef CYL_COEFS      // This is the one you get if you don't #define any
#undef SQUARE_COEFS
#define PAPER_COEFS

// Miscellaneous #define's
#undef OLD_CODE
#undef VERBOSE
#undef PRECESSION   // Define this to turn on precession term
#define PRECESSION_RATIO (-1./Grid2D::DampCoef)
          // Relative size of precession term. See above discussion
          // about DampCoef and GyRatio.

#undef CODE_CHECKS

// ODE Solvers; Define exactly one of the following
#undef EULER_ODE       // 1st order Euler
#define PREDICTOR_2_5_ODE // 2.5 order predictor
#undef PREDICTOR_CORRECTOR_ODE // 2nd order predictor corrector
#undef RUNGE_KUTTA_ODE  // 4th order Runge-Kutta
                        // (cf. MaxTorqueStep note below)
#undef RUNGE_KUTTA_FAST // Define this *in addition to RUNGE_KUTTA_ODE*
                         // to use approximate field values (i.e.,
                         // hFastUpdate) for internal ODE steps.

// Error control
#ifndef RUNGE_KUTTA_ODE
const double MaxTorqueStep  =  0.09; // StepSize*maxtorque restricted to
     // being less than this amount.  This will roughly limit magnetization
     // shifts to < atan(0.09) --- about 5 degrees.
#else  // Current Runge-Kutta solver takes double step
const double MaxTorqueStep  =  0.18;
#endif
const double AllowedSpinError =  1e-3;  // Allowed error per spin

#ifdef STOPWATCH_ON
#define STOPWATCH_START(watch) watch.Start();
#define STOPWATCH_STOP(watch)  watch.Stop();
#else // !STOPWATCH_ON
#define STOPWATCH_START(watch)
#define STOPWATCH_STOP(watch)
#endif // STOPWATCH_ON

// Grid (mesh) file magic cookie string
// const char GridFileMagic03[8] = "G2DFil\x03";   OBSOLETE
// const double GridFileEndianTest = 1.234;        OBSOLETE
// The bit pattern for 1.234 on a little endian machine (e.g.,
// Intel's x86 series) should be 58 39 B4 C8 76 BE F3 3F, while
// on a big endian machine it is 3F F3 BE 76 C8 B4 39 58.  Examples
// of big endian machines include ones using the MIPS R4000 and R8000
// series chips, for example SGI's.


/////////////////////////// Grid2D Code /////////////////////////////////
// Grid I/O macros
#define WRITEVAR(x,fptr) fwrite(&(x),sizeof(x),1,fptr)
#define READVAR(x,fptr,swap) \
  ( fread((char *)(&(x)),sizeof(x),1,fptr), (swap) ? (x)=flip(x) : (x) )

double Grid2D::GetTimeStep(void)
{ // Returns size of last step, in seconds
  return StepSize/(GyRatio*DampCoef*Ms);
}

void Grid2D::Allocate(int xcount,int zcount)
{ // NOTE: This routine requires that Grid2d::ExternalDemag be initialized,
  //       since *ExternalDemag->init is called.
  int i;

  Nx=xcount; Nz=zcount;
  if(DoDemag) ThreeVector::Alloc2DArray(Nx,Nz,hdemag);
  else        hdemag=(ThreeVector**)NULL;
  ThreeVector::Alloc2DArray(Nx,Nz,h);
  ThreeVector::Alloc2DArray(Nx,Nz,torque);
  ThreeVector::Alloc2DArray(Nx,Nz,torque0);
  ThreeVector::Alloc2DArray(Nx,Nz,torque1);
  ThreeVector::Alloc2DArray(Nx,Nz,h0);
  ThreeVector::Alloc2DArray(Nx,Nz,h1);

  // Allocate MagElt arrays
  m0.Allocate(Nx,Nz);
  m1.Allocate(Nx,Nz);
  m2.Allocate(Nx,Nz);
  m.Allocate(Nx,Nz);
  moment_sum=Nx*Nz;  // Note: This will be recalculated after
  /// the sample geometry is set.

  // Allocate demag calculation arrays
  if(DoDemag) {
    if(ExternalDemag->calc==NULL) { // Default to internal routine
      // Allocate demagcoef matrices
      if(!(ADemagCoef=new double*[Nx]))
	PlainError(1,"Error in Grid2D::Allocate: %s",ErrNoMem);
      if(!(CDemagCoef=new double*[Nx]))
	PlainError(1,"Error in Grid2D::Allocate: %s",ErrNoMem);
      for(i=0;i<Nx;i++) {
	if(!(ADemagCoef[i]=new double[Nz]))
	  PlainError(1,"Error in Grid2D::Allocate: %s",ErrNoMem);
	if(!(CDemagCoef[i]=new double[Nz]))
	  PlainError(1,"Error in Grid2D::Allocate: %s",ErrNoMem);
      }
    }
    else { // Use external routine
      (*ExternalDemag->init)(Nx,Nz,ExternalDemag->paramcount,
			     ExternalDemag->params);
    }
  } // fi(DoDemag)
}


int Grid2D::FindNeighbors(int nsize,Array2D<MagElt> &me,MagEltLink men[],
			       int i,int k)
{ // NOTE: Currently does only 4 and 8 neighborhoods.
  // NOTE 2: Enforces Neuman bdry conditions along edges of gridded
  //       region.  Hmm, you may get different results if the sample edge
  //       is not the same as the gridded region edge (e.g., by setting
  //       |m|=0 inside gridded region to, for example, round corners.)
  //       -mjd 5-Dec-1996.
#ifdef CODE_CHECKS
  if(nsize!=4 && nsize!=8)
    PlainError(1,"Error in Grid2D::FindNeighbors: "
	     "Unsupported neighborhood size (%d) request",nsize);
  if(me.GetSize()<1 || men==(MagEltLink*)NULL) 
    PlainError(1,"Error in Grid2D::FindNeighbors: Null pointer received");
  if(i<0 || i>=Nx || k<0 || k>=Nz)
    PlainError(1,"Error in Grid2D::FindNeighbors: "
	     "Requested index elt (%d,%d) out-of-bounds (0-%d,0-%d)",
	     i,k,Nx,Nz);
#endif
  hValid=torqueValid=energyValid=0;
  if(me[i][k].GetMoment()<=0.) return 0; // Zero moment elements
  /// don't get connected.

  static int ioff[8]={ 1, 1, 0,-1,-1,-1, 0, 1};
  static int koff[8]={ 0, 1, 1, 1, 0,-1,-1,-1};
  int n,offset,offstep,it,kt,new_ngbrcnt;
  if(nsize==4) offstep=2;  else offstep=1;
  for(n=offset=new_ngbrcnt=0;n<nsize;n++,offset+=offstep) {
    it=i+ioff[offset];
    kt=k+koff[offset];

    // Enforce Neumann bdry conds.,  \part(m)/\part(n)=0,
    // at grid edges
    if(it<0)        it=0;
    else if(it>=Nx) it=Nx-1;
    if(kt<0)        kt=0;
    else if(kt>=Nz) kt=Nz-1;
    if(it==i && kt==k) continue;

    // Handle irregular boundaries
    if(me[it][kt].GetMoment()<=0.) {
      if(it==i || kt==k) continue;  // Set up alternate connection
      /// only for diagonal links

      // See which of the nearest neighbors is non-zero
      int itt=it;  int ktt=kt;
      if(me[i][kt].GetMoment()>0.) itt=i;
      if(me[it][k].GetMoment()>0.) ktt=k;

      if(itt==it && ktt==kt) continue; // Neither; Skip link
      if(itt==i  && ktt==k)  continue; // Both; Skip link

      it=itt; kt=ktt;
    }

#ifdef CODE_CHECKS
    if((it==i && kt==k) || me[it][kt].GetMoment()<=0.)
      PlainError(1,"Error in Grid2D::FindNeighbors: "
	       "Programming error: Bad neighbor link");
#endif // CODE_CHECKS

    // Add neighbor
    men[new_ngbrcnt].weight=1.0;
    men[new_ngbrcnt++].node=&(me[it][kt]);    
  }
  return new_ngbrcnt;
}
#ifndef UNIT_MAG
void Grid2D::InitRectangle(String &)
{ // Default part geometry
  return; // Nothing to do
}
void Grid2D::InitEllipse(String &)
{ // Zero's moments outside ellipse defined by Nx,Nz
  if(Nx<2 || Nz<2) return; // Nothing to do
  int i,k; double x,z;
  for(i=0;i<Nx;i++) {
    x= double(2*i)/double(Nx-1)-1.0;
    for(k=0;k<Nz;k++) {
      z= double(2*k)/double(Nz-1)-1.0;
      if(SQ(x)+SQ(z)>1+SQRT_EPSILON) m[i][k].SetMoment(0.0);
    }
  }
  hValid=torqueValid=energyValid=0;
}

void Grid2D::Mask(String &filepoint)
{ // Zeros moments if not black
#define Maxline 256 
  static char buf[Maxline];
  int i,k,hapval=0,ppmtype=0;
  FILE* mp;
  Omf_TclNativeFilename(filepoint);
/**/ fprintf(stderr,"Trying to open file %s\n",filepoint.GetStr());
  mp=Omf_FOpen(filepoint.GetStr(),"rb");
/**/ fprintf(stderr,"Omf_FOpen return value: %p\n",mp);
  if(mp==NULL) {
    PlainError(1,"Error in Grid2D::Mask: Unable to open file: %s",
	       filepoint.GetStr());
  }
  while(hapval<=2) {    
    if(fgets(buf,Maxline,mp)==NULL)
      PlainError(1,"Error in Grid2D::Mask: Error reading file: %s",
		 filepoint.GetStr());
    if((strstr(buf,"#")==NULL)&&(strspn(buf," \t\n")!=strlen(buf)))  {
      if((hapval==0)&&(strstr(buf,"P3")!=NULL)) ppmtype=3;
      else if ((hapval==0) &&
	       ((strstr(buf,"P6")!=NULL)|| (strstr(buf,"P4")!=NULL)))
	ppmtype=6;
      else if (hapval==0)
	PlainError(1,"Error in Grid2D::Mask: Not a valid ppm file: %s",
		   filepoint.GetStr());
      hapval++;
    } 
  }
  if (ppmtype==3) {
    int Red,Green,Blue;
    for(k=Nz-1;k>=0;k--)  for(i=0;i<Nx;i++){
      if (fscanf(mp,"%i %i %i",&Red,&Green,&Blue)==EOF)
	PlainError(1,"Error in Grid2D::Mask(String filepoint) "
		 "PPM size does not match grid size in: %s",
		 filepoint.GetStr());
      else
        if(Red+Green+Blue) m[i][k].SetMoment(0.0); 
    }
  }
  else {
    int Check=0,*intbuf;
    intbuf=&Check;
    for(k=Nz-1;k>=0;k--)  for(i=0;i<Nx;i++){
      if(feof(mp))
        PlainError(1,"Error in Grid2D::Mask(String filepoint) "
		   "PPM size does not match grid size in: %s",
		   filepoint.GetStr());
      else {
	fread((char *)intbuf,3,1,mp);
	if(Check) m[i][k].SetMoment(0.0);    
      }
    }
  }
  fclose(mp);
}

void Grid2D::InitOval(String &inprop)
{ // Rounds corners off Nx by Nz rectangle by zeroing
  // moments outside of quarter circles of radius...
  double rounded_proportion=atof(inprop.GetStr());
  
  double radius=0.5*rounded_proportion*MIN(Nx,Nz);

  // Corner rounding circle center coordinates
  double cx1,cx2,cz1,cz2;
  cx1=radius-0.5;   cx2=Nx-0.5-radius;
  cz1=radius-0.5;   cz2=Nz-0.5-radius;
  if(cx1>cx2 || cz1>cz2)
    PlainError(1,"Error in Grid2D::InitOval: %s",ErrProgramming);

  // Round lower left corner
  int i,k;   double xdistsq,zdistsq,radsq=SQ(radius);
  for(i=0;i<(int)floor(cx1);i++) {
    xdistsq=cx1-i; xdistsq*=xdistsq;
    for(k=0;k<(int)floor(cz1);k++) {
      zdistsq=cz1-k; zdistsq*=zdistsq;
      if(xdistsq+zdistsq>radsq+SQRT_EPSILON) m[i][k].SetMoment(0.0);
    }
  }
  // Round lower right corner
  for(i=(int)ceil(cx2);i<Nx;i++) {
    xdistsq=cx2-i; xdistsq*=xdistsq;
    for(k=0;k<(int)floor(cz1);k++) {
      zdistsq=cz1-k; zdistsq*=zdistsq;
      if(xdistsq+zdistsq>radsq+SQRT_EPSILON) m[i][k].SetMoment(0.0);
    }
  }
  // Round upper right corner
  for(i=(int)ceil(cx2);i<Nx;i++) {
    xdistsq=cx2-i; xdistsq*=xdistsq;
    for(k=(int)floor(cz2);k<Nz;k++) {
      zdistsq=cz2-k; zdistsq*=zdistsq;
      if(xdistsq+zdistsq>radsq+SQRT_EPSILON) m[i][k].SetMoment(0.0);
    }
  }
  // Round upper left corner
  for(i=0;i<(int)floor(cx1);i++) {
    xdistsq=cx1-i; xdistsq*=xdistsq;
    for(k=(int)floor(cz2);k<Nz;k++) {
      zdistsq=cz2-k; zdistsq*=zdistsq;
      if(xdistsq+zdistsq>radsq+SQRT_EPSILON) m[i][k].SetMoment(0.0);
    }
  }

  hValid=torqueValid=energyValid=0;
}

void Grid2D::SetBoundaryMoments(double momentsize)
{ // Sets all boundary magelt's moment magnitudes to momentsize
  // Note: This subroutine will likely change which elements
  //       answer "1" to MagElt::IsBoundary().
  int i,k,bdrycount=0;
  MagElt** meptr=new MagElt *[Nx*Nz];
  if(!meptr) 
    PlainError(1,"Error in Grid2D::SetBoundaryMoments(double): %s",ErrNoMem);
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    if(m[i][k].GetMoment()>0. && m[i][k].IsBoundary())

      meptr[bdrycount++]=&m[i][k];
  }
  for(i=0;i<bdrycount;i++) meptr[i]->SetMoment(momentsize);
  delete meptr;
  hValid=torqueValid=energyValid=0;
}

void Grid2D::RoughenBoundaryMoments()
{ // Multiplies each boundary moment (separately) by an amount
  // drawn from a uniform distribution on [0,1].
  int i,k,bdrycount=0;
  MagElt** meptr=new MagElt *[Nx*Nz];
  if(!meptr) {
    PlainError(1,"Error in Grid2D::RoughenBoundaryMoments(double): %s",
	       ErrNoMem);
  }
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    if(m[i][k].GetMoment()>0. && m[i][k].IsBoundary())
      meptr[bdrycount++]=&m[i][k];
  }
  for(i=0;i<bdrycount;i++) {
    meptr[i]->SetMoment(meptr[i]->GetMoment()*UnifRand());
  }
  delete meptr;
  hValid=torqueValid=energyValid=0;
}

#endif // UNIT_MAG


void Grid2D::Initialize(MIF &mif,double new_mesh_size,
			MagEltAnisType new_magelt_anistype)
{
  double new_Ms=mif.GetMs();
  double new_K1=mif.GetK1();
  double new_exchange_stiffness=mif.GetA();
  double new_stepsize=mif.GetInitialIncrement();

  int i,k;
  double dtemp;

  energy=1.0/EPSILON;  // Just initialize to arbitrarily big number
  exch_energy=anis_energy=demag_energy=zeeman_energy=energy;

  Ms                 = new_Ms;
  K1                 = new_K1;
  exchange_stiffness = new_exchange_stiffness;
  mesh_size          = new_mesh_size;
  magelt_anistype    = new_magelt_anistype;
  MagElt::SetupConstants(new_Ms,new_exchange_stiffness,
			 new_mesh_size,new_magelt_anistype);
  StepSize=MAX(EPSILON,new_stepsize);
  hUpdateCount=0; // Count of calls to hUpdate
  ODEIterCount=0; // Count of calls to hUpdate
  hValid=torqueValid=energyValid=0;

  // Setup internal demag function, if necessary
  // (see also Grid2D::Allocate())
  if(ExternalDemag->calc==(DF_Calc *)NULL && DoDemag) {
    ADemagCoef[0][0]=CDemagCoef[0][0]=0.0;
#if defined PAPER_COEFS
    SelfDemag= -0.5;
#elif defined SQUARE_COEFS
    SelfDemag= -0.5;
#else // CYL_COEFS
    SelfDemag=(-PI/6);
#endif
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
      if(i!=0 || k!=0) {
#if defined PAPER_COEFS // Use MSZ coefficients (APS Vol 49, #22, p15745-15752)
	if((dtemp=SQ(2*i+1)+4*SQ(k))<0)
	  PlainError(1,"Error in Grid2D::Initialize: %s","Integer overflow");
	ADemagCoef[i][k]=atan2(4*i-2,SQ(2*i-1)+4*SQ(k)-1)
	  -atan2(4*i+2,SQ(2*i+1)+4*SQ(k)-1);
	dtemp=4*SQ(SQ(i))+8*SQ(i)*SQ(k)+4*SQ(SQ(k))+1;
	CDemagCoef[i][k]=0.5*log((dtemp+8*i*k)/(dtemp-8*i*k));
	ADemagCoef[i][k]/=2*PI; // I don't know if this is the proper
	CDemagCoef[i][k]/=2*PI; // normalization or not, but it looks good.
#elif defined SQUARE_COEFS
	dtemp=(2*k+1)/sqrt(SQ(2*i-1)+SQ(2*k+1));
	dtemp-=(2*k-1)/sqrt(SQ(2*i-1)+SQ(2*k-1));
	ADemagCoef[i][k]=2*dtemp/(2*i-1);
	dtemp=(2*k+1)/sqrt(SQ(2*i+1)+SQ(2*k+1));
	dtemp-=(2*k-1)/sqrt(SQ(2*i+1)+SQ(2*k-1));
	ADemagCoef[i][k]-=2*dtemp/(2*i+1);
	CDemagCoef[i][k]=2./sqrt(SQ(2*i-1)+SQ(2*k-1))
	                -2./sqrt(SQ(2*i-1)+SQ(2*k+1))
	                -2./sqrt(SQ(2*i+1)+SQ(2*k-1))
	                +2./sqrt(SQ(2*i+1)+SQ(2*k+1));
#else // CYL_COEFS
	if((dtemp=(i*i+k*k))<1)
	  PlainError(1,"Error in Grid2D::Initialize: Integer overflow");
	dtemp=2*PI*SQ(dtemp); // Denominator (Area consider. cancel mesh_size)
	ADemagCoef[i][k]=(i*i-k*k)/dtemp;
	CDemagCoef[i][k]=2.0*(i*k)/dtemp;
#endif // Coefficient selection
      }
    }
  }

  // Set initial H field to 0
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    h[i][k][0]=h[i][k][1]=h[i][k][2]=0.0;
    h0[i][k][0]=h0[i][k][1]=h0[i][k][2]=0.0;
    h1[i][k][0]=h1[i][k][1]=h1[i][k][2]=0.0;
    if(DoDemag) hdemag[i][k][0]=hdemag[i][k][1]=hdemag[i][k][2]=0.0;
  }

  // Set initial torques to 0
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    torque1[i][k][0]=torque1[i][k][1]=torque1[i][k][2]=0.0;
    torque0[i][k][0]=torque0[i][k][1]=torque0[i][k][2]=0.0;
    torque[i][k][0]=torque[i][k][1]=torque[i][k][2]=0.0;
  }

  hValid=torqueValid=energyValid=0;  // Safety;  Should already be set to 0 above
}

Grid2D::Grid2D(MIF &mif):ZeroField(0,NULL)
{
  if(!mif.IsValid()) {
    PlainError(1,"Error in Grid2D::Grid2D: Invalid mif object.");
  }
  dfparams=(char **)NULL; // Safety;
  UnifRand(mif.GetRandomizerSeed());
  double zeroarr[]={0.,0.,0.};
  ZeroField.SetNomField(mif.GetMs(),ThreeVector(zeroarr));
  SethApplied(&ZeroField);  // Safety

  int xcount,zcount,i,k;
  MagEltAnisType new_magelt_anistype=meat_uniaxial; // Safety
  moment_sum=1.0;  // Safety

  // Get basic info
  if((cellsize=mif.GetCellSize())<=0.0)
    PlainError(1,"Error in Grid2D::Grid2D(MIF &): "
	     "Illegal cell size (%f meters)",cellsize);
  xcount=(int)rint(mif.GetPartWidth()/cellsize);
  zcount=(int)rint(mif.GetPartHeight()/cellsize);
  xmax=xcount*cellsize;
  zmax=zcount*cellsize;
  ymax=mif.GetPartThickness();

  // Do demag initialization
  if(!(dfparams=new char*[MAX(2,MAXDFPARAMS)])
     || !(dfparams[0]=new char[MAXDFNAMELENGTH]))
    PlainError(1,"Error in Grid2D::Grid2D(MIF &): %s",ErrNoMem);
  strncpy(dfparams[0],mif.GetDemagName(),MAXDFNAMELENGTH);
  *(dfparams[1]=(char *)strchr((const char *)dfparams[0],'\0')+1)='\0';
  if(DF_LookUpName(dfparams[0],&ExternalDemag)!=0) // No match!
    PlainError(1,"Error in Grid2D::Grid2D(MIF &): "
	     "No demag calculation routine by the name ->%s<- found",
	     dfparams[0]);
  if(Omf_StrCaseCmp(ExternalDemag->name,"NoDemag")==0) DoDemag=0;
  else                                                 DoDemag=1;
  if(ExternalDemag->paramcount<1) ExternalDemag->params=(char**)NULL;
  else {
    ExternalDemag->params=dfparams;
    if(ExternalDemag->paramcount>1)
      sprintf(dfparams[1],"%.15g",mif.GetPartThickness()/cellsize);
  }

  // Determine magelt_anistype and anisotropy directions
#if (VECDIM != MIFVECDIM)
#error VECDIM and MIFVECDIM must be equal
#endif
  switch(mif.GetAnisType())
    {
    case mif_uniaxial: new_magelt_anistype=meat_uniaxial; break;
    case mif_cubic:    new_magelt_anistype=meat_gencubic; break;
    default: PlainError(1,"Error in Grid2D::Grid2D(MIF &): "
			"Unknown Anisotropy Type in MIF");
	     break;
    }
  double dirA_arr[VECDIM];  ThreeVector dirA(dirA_arr);
  double dirB_arr[VECDIM];  ThreeVector dirB(dirB_arr);
  double dirC_arr[VECDIM];  ThreeVector dirC(dirC_arr);
  dirA.Dup(mif.GetAnisDir(1));
  if(fabs(dirA.MagSq()-1)>SQRT_EPSILON)
    PlainError(1,"Error in Grid2D::Grid2D(MIF &): Non-unit anisotropy direction (1).");
  if(new_magelt_anistype!=meat_uniaxial) { // Some cubic type
    dirB.Dup(mif.GetAnisDir(2));
    if(fabs(dirB.MagSq()-1)>SQRT_EPSILON)
      PlainError(1,"Error in Grid2D::Grid2D(MIF &): Non-unit anisotropy direction (2).");
    double dot=dirA*dirB;
    if(fabs(dot)>SQRT_EPSILON)
      PlainError(1,"Error in Grid2D::Grid2D(MIF &): Crystal axes not perpendicular.");
    dirC=dirA^dirB;
    if(fabs(dirC.MagSq()-1)>SQRT_EPSILON)
      PlainError(1,"Error in Grid2D::Grid2D(MIF &): Non-unit anisotropy direction (3).");
    if(dirA[0]==1.0 && dirB[1]==1.0)
      new_magelt_anistype=meat_cubic;  // Standard cubic anisotropy
  }

  // Allocate arrays
  Allocate(xcount,zcount);

  // Setup applied field pointer.  This must be done *after*
  // Nx & Nz are setup (which is currently done in Allocate).
  // NOTE: *hApplied is owned external to Grid2D
  SethApplied(mif.GetExternalFieldObject());
  mif.ResetField();

  // Initialize program constants
  Initialize(mif,cellsize,new_magelt_anistype);

  // Setup anisotropy axes inside MagElt's.  Anisotropy arrays should
  // be unchanging, so they can be shared across all the magelt arrays.
  ANIS_INIT_FUNC* aifp=AnisInitList.GetAnisInitFunc(mif.GetAnisInitName());
  if(aifp==NULL)
    PlainError(1,"Error in Grid2D::Grid2D(MIF &): "
	       "Unknown AnisInitFunc: ->%s<-",mif.GetAnisInitName());
  (*aifp)(mif.GetAnisInitArgc(),mif.GetAnisInitArgv(),
	  Nx,Nz,magelt_anistype,K1,dirA,dirB,dirC,
	  m,anisa,anisb,anisc);
  double *aptr=NULL,*bptr=NULL,*cptr=NULL;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m[i][k].GetAnisDirs(aptr,bptr,cptr);   // Copy anis pointers from m
    m0[i][k].InitAnisDirs(aptr,bptr,cptr); // to m0,
    m1[i][k].InitAnisDirs(aptr,bptr,cptr); // to m1,
    m2[i][k].InitAnisDirs(aptr,bptr,cptr); // and to m2,
  }

  // Set up pointers for m initialization routine
  char **magptr=mif.GetMagInitArgv();
  InitM=MagInit::LookUp(magptr[0]);
  InitMParams=magptr+1;

#ifndef UNIT_MAG
  // Setup sample geometry.  NOTE: We need to setup the geometry (by
  //  setting some magelt moments to 0) _before_ calling the neighborhood
  //  constructor routine, because the latter needs to know where the
  //  boundaries are in order to get proper boundary conditions.
  // NOTE 2: These geometry initialization routines must not
  //  use the m->ngbr arrays (obviously).
   MifGeometryType mgt; String geomparam;
   mgt=mif.GetGeometry(geomparam);
  if(mgt==mif_rectangle)    InitRectangle(geomparam);
  else if(mgt==mif_ellipse) InitEllipse(geomparam);
  else if(mgt==mif_oval)    InitOval(geomparam);
  else if(mgt==mif_mask)    Mask(geomparam);
  else PlainError(1,"Error in Grid2D::Grid2D(MIF &): Part geometry not specified.");
#ifdef WEAK_BOUNDARY
  if(WEAK_BOUNDARY>0.0) SetBoundaryMoments(WEAK_BOUNDARY);
#endif // WEAK_BOUNDARY
#ifdef ROUGH_BOUNDARY
  RoughenBoundaryMoments();
#endif // ROUGH_BOUNDARY
#endif // !UNIT_MAG

  // Initialize m and copy data (vector, moment and K1) into m0, m1,
  // and m2.  Note that _after_ this all magelt's have their moment
  // magnitude set, so then one may set up the nbhd structure.
  if(InitM!=NULL) // InitM==NULL => m already setup
    (*InitM)(Nx,Nz,m,InitMParams);
  moment_sum=0.0;
  for(i=0;i<Nx;i++) {
    double moment_column_sum=0.0;
    for(k=0;k<Nz;k++) {
      moment_column_sum+=fabs(m[i][k].GetMoment());
      m0[i][k].CopyData(m[i][k]);
      m1[i][k].CopyData(m[i][k]);
      m2[i][k].CopyData(m[i][k]);
    }
    moment_sum+=moment_column_sum;
  }
  if(moment_sum<EPSILON) moment_sum=EPSILON;  // Safety

  // Set up neighbor structure in m and m0, m1, m2
  int new_ngbrcnt;
  int maxmagngbrs=MagElt::GetMaxNgbrcnt();
  MagEltLink* Neighbors = new MagEltLink[maxmagngbrs];

  if(maxmagngbrs!=4 && maxmagngbrs!=8)
   PlainError(1,"Error in Grid2D::Grid2D(MIF &): "
	    "Current code only supports 4 and 8 neighborhoods");

  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    new_ngbrcnt=FindNeighbors(maxmagngbrs,m,Neighbors,i,k);
    m[i][k].SetupNeighbors(new_ngbrcnt,Neighbors);
    new_ngbrcnt=FindNeighbors(maxmagngbrs,m0,Neighbors,i,k);
    m0[i][k].SetupNeighbors(new_ngbrcnt,Neighbors);
    new_ngbrcnt=FindNeighbors(maxmagngbrs,m1,Neighbors,i,k);
    m1[i][k].SetupNeighbors(new_ngbrcnt,Neighbors);
    new_ngbrcnt=FindNeighbors(maxmagngbrs,m2,Neighbors,i,k);
    m2[i][k].SetupNeighbors(new_ngbrcnt,Neighbors);
  }
  delete[] Neighbors;

}

void Grid2D::Deallocate(void)
{
  int i;
  if(ExternalDemag->calc!=NULL) (*ExternalDemag->destroy)();
  ThreeVector::Release2DArray(Nx,h);
  ThreeVector::Release2DArray(Nx,h0);
  ThreeVector::Release2DArray(Nx,h1);
  ThreeVector::Release2DArray(Nx,anisa);
  ThreeVector::Release2DArray(Nx,anisb);
  ThreeVector::Release2DArray(Nx,anisc);
  if(DoDemag) ThreeVector::Release2DArray(Nx,hdemag);
  ThreeVector::Release2DArray(Nx,torque1);
  ThreeVector::Release2DArray(Nx,torque0);
  ThreeVector::Release2DArray(Nx,torque);
  m.Deallocate(); m0.Deallocate(); m1.Deallocate(); m2.Deallocate();
  for(i=0;i<Nx;i++) {
    if(ExternalDemag->calc==NULL && DoDemag)
      { delete[] ADemagCoef[i]; delete[] CDemagCoef[i]; }
  }
  if(ExternalDemag->calc==NULL && DoDemag)
    { delete[] ADemagCoef; delete[] CDemagCoef; }
  if(dfparams!=NULL) { delete dfparams[0]; delete dfparams; }
}

Grid2D::~Grid2D(void)
{
  Deallocate();
}

void Grid2D::PrintUsageTimes(FILE *fptr)
{
#ifdef STOPWATCH_ON
  fprintf(fptr,"# Exchange time---\n");   sw_exch.Print(fptr);
  fprintf(fptr,"#\n# Anisotropy time---\n"); sw_anis.Print(fptr);
  fprintf(fptr,"#\n# Demag time---\n");      sw_demag.Print(fptr);
  fflush(fptr);
#endif // STOPWATCH_ON
}

void Grid2D::Perturb(double max_mag)
{ // Randomly perturb all m by <magnitude
  for(int i=0;i<Nx;i++) for(int k=0;k<Nz;k++) {
    m[i][k].Perturb(max_mag);
  }
  hValid=torqueValid=energyValid=0;
}

// The next two functions (ADemag() & CDemag()) access the demagcoef
// matrices using symmetries to allow negative values for i and k.
double Grid2D::ADemag(int i,int k)
{
#ifdef OLD_CODE
  return ADemagCoef[abs(i)][abs(k)];
#else
  return ADemagCoef[(i>=0?i:-i)][(k>=0?k:-k)];
#endif
}

double Grid2D::CDemag(int i,int k)
{
#ifdef OLD_CODE
  if(i==0 || k==0) return 0.0;
  double dtemp=CDemagCoef[abs(i)][abs(k)];
  if((i>0 && k>0) || (i<0 && k<0)) return  dtemp;
  else                             return -dtemp;
#else
  int sign=1;
  if(i<0) { i*=-1; sign=-1; }
  if(k<0) { k*=-1; sign*=-1; }
  return  sign*CDemagCoef[i][k];
#endif
}

/* Calculate the demag field at (i,k) including the self-field */
ThreeVector& Grid2D::CalculateDemag(int i,int k)
{
  static double hdarr[3];
  static ThreeVector hd(hdarr);
  if(!DoDemag) return hd;

  hd[0]=hd[1]=hd[2]=0.0;

  double A,C,mx,mz;
#ifndef UNIT_MAG
  double mom;
#endif
  int i2,k2;
  for(i2=0;i2<Nx;i2++) for(k2=0;k2<Nz;k2++) {
    A=ADemag(i2-i,k2-k); C=CDemag(i2-i,k2-k);
    mx=m[i2][k2][0];     mz=m[i2][k2][2];
#ifndef UNIT_MAG
    hd[0]+=(mom=m[i2][k2].GetMoment())*(A*mx+C*mz);
    hd[2]+=mom*(C*mx-A*mz);
#else
    hd[0]+=A*mx+C*mz;    hd[2]+=C*mx-A*mz;
#endif
  }
#ifndef UNIT_MAG
  mom=m[i][k].GetMoment();
  hd[0]+=mom*SelfDemag*m[i][k][0]; // Self-demag
  hd[2]+=mom*SelfDemag*m[i][k][2];
#else
  hd[0]+=SelfDemag*m[i][k][0]; // Self-demag
  hd[2]+=SelfDemag*m[i][k][2];
#endif
  return hd;
} 

/* Calculate the energy for the entire mesh */
void Grid2D::CalculateEnergy(void)
{
  int i,k;
  double exchcol,aniscol,demagcol,zeemancol;
  MagElt* mptr;
  double happmag,happarr[VECDIM]; ThreeVector happ(happarr);

  if(!hValid) hUpdate();

  exch_energy=anis_energy=demag_energy=zeeman_energy=0.0;
  exchcol=aniscol=demagcol=zeemancol=0.0;  // Just to shut up compiler

  for(i=0;i<Nx;i++) {
    exchcol=aniscol=demagcol=zeemancol=0.0;
    STOPWATCH_START(sw_exch);
    for(k=0,mptr=m[i];k<Nz;k++,mptr++)
      exchcol+=mptr->CalculateExchangeEnergy();
    STOPWATCH_STOP(sw_exch);
    STOPWATCH_START(sw_anis);
    for(k=0,mptr=m[i];k<Nz;k++,mptr++)
      aniscol+=(mptr->*MagElt::AnisotropyEnergy)();
    STOPWATCH_STOP(sw_anis);
    STOPWATCH_START(sw_demag);
    for(k=0,mptr=m[i];k<Nz;k++,mptr++) {
      if(DoDemag) {
#ifndef UNIT_MAG
	demagcol+= mptr->GetMoment() * ((*mptr)*hdemag[i][k]);
	           // *(-1); see note (**) below
#else
	demagcol+=(m[i][k]*hdemag[i][k]);  // *(-1); see note (**) below
#endif // UNIT_MAG
      }
    }
    STOPWATCH_STOP(sw_demag);
    for(k=0,mptr=m[i];k<Nz;k++,mptr++) {
      hApplied->GetLocalhField(i,k,happ);
      happmag=hApplied->GetLocalhFieldMag(i,k);
#ifndef UNIT_MAG
      zeemancol+=mptr->GetMoment()*(happmag-((*mptr)*happ));
#else
      zeemancol+=happmag-((*mptr)*happ);
#endif // UNIT_MAG
      // Zeeman is offset (by hFieldMag) so minimum energy is 0
    }
    // Accumulate by columns to improve precision
    exch_energy+=exchcol;      anis_energy+=aniscol;
    demag_energy+=demagcol;  zeeman_energy+=zeemancol;
  }
  // demag_energy gets divided by 2 because changing m also changes h.
  demag_energy/=2;

  // Normalize energies
  exch_energy/=moment_sum;
  anis_energy/=moment_sum;
  demag_energy/=(-moment_sum);    // Demag above off by minus sign  (**)
  zeeman_energy/=moment_sum;

  // Total energy density---
  energy = exch_energy+anis_energy+demag_energy+zeeman_energy;
  energyValid=1;
}

void Grid2D::InternalDemagCalc(void)
{
#ifdef CODE_CHECKS
  if(!DoDemag)
    PlainError(1,"Error in Grid2D::InternalDemagCalc: "
	     "This routine shouldn't be called because DoDemag is false");
#endif
  for(int i=0;i<Nx;i++) for(int k=0;k<Nz;k++)
    hdemag[i][k]=CalculateDemag(i,k);
}

void Grid2D::hUpdate(void)
{
#ifdef CODE_CHECKS
  if(hValid) PlainWarning("Warning in Grid2D::hUpdate: "
			  "Updating h even though hValid is TRUE.");
#endif // CODE_CHECKS
  int i,k;
  hUpdateCount++;

  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) hApplied->GetLocalhField(i,k,h[i][k]);
  STOPWATCH_START(sw_exch);
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) h[i][k]+=m[i][k].CalculateExchange();
  STOPWATCH_STOP(sw_exch);
  STOPWATCH_START(sw_anis);
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
    h[i][k]+=(m[i][k].*MagElt::AnisotropyField)();
  STOPWATCH_STOP(sw_anis);
  if(DoDemag) {
    STOPWATCH_START(sw_demag);
    if(ExternalDemag->calc!=NULL) // Use external demag routine
      (*ExternalDemag->calc)(WriteGrid2Dm,FillGrid2Dh,m,hdemag);
    else // Use internal explicit (brute-force) demag calculation
      InternalDemagCalc();
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) h[i][k]+=hdemag[i][k];
    STOPWATCH_STOP(sw_demag);
  }

  hValid=1;
  torqueValid=energyValid=0;
}

void Grid2D::hFastUpdate(void)
{
  int i,k;

  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) hApplied->GetLocalhField(i,k,h[i][k]);
  STOPWATCH_START(sw_exch);
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) h[i][k]+=m[i][k].CalculateExchange();
  STOPWATCH_STOP(sw_exch);
  STOPWATCH_START(sw_anis);
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
    h[i][k]+=(m[i][k].*MagElt::AnisotropyField)();
  STOPWATCH_STOP(sw_anis);
  if(DoDemag) {
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) h[i][k]+=hdemag[i][k];
  }                                           // Use old demag
  torqueValid=energyValid=0;
}

void Grid2D::CalculateTorque(Array2D<MagElt> &m_in,
			       ThreeVector** &h_in,
			       ThreeVector** &t_out)
{ // Calculates torque from imports m_in & h_in, storing
  // result in t_out.
  static double vtemparr[3];  static ThreeVector vtemp(vtemparr);
  int i,k;
  MagElt *mp;
  ThreeVector *hp;
  ThreeVector *tp;

  for(i=0;i<Nx;i++) {
    hp=h_in[i]; tp=t_out[i]; mp=m_in[i];
    for(k=0;k<Nz;k++,hp++,tp++,mp++) {
      vtemp= *mp;
      tp->FastAdd(*hp,-(vtemp*(*hp)),vtemp);
#ifdef PRECESSION
      vtemp^=(*hp);                 // Precession term
      tp->Accum(PRECESSION_RATIO,vtemp);
#endif // PRECESSION

#ifndef UNIT_MAG
      (*tp) *= mp->GetMoment();
#endif
    }
  }
}

double Grid2D::GetMaxTorque(void)
{
  int i,k;
  double dtemp,maxtorque=0.0;
  if(!hValid) hUpdate();
  if(!torqueValid) { CalculateTorque(m,h,torque); torqueValid=1; }
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    if((dtemp=torque[i][k].MagSq())>maxtorque) maxtorque=dtemp;
  }
  return sqrt(maxtorque);
}

void Grid2D::StepEuler(double &next_stepsize)
{ // Takes one Euler step, with stepsize control based on energy.  Assumes
  // m, h, torque & energy are current coming in, and leaves them current
  // going out.  Saves entry values of m, h, and torque in m0, h0, and
  // torque0.
  //   The attempted step size is determined by the value of *this->StepSize
  // on entry.  StepSize may be adjusted by this routine; the value of
  // StepSize on exit is the step size actually used.  The export
  // next_stepsize should be used by the calling routine (on the next pass)
  // to set *this->StepSize _before_ calling this routine.

  const double CutRatio=0.5;        // Determined by trial & error
  const double IncreaseRatio=1.1;   //             ""
  // NOTE: I've tried more sophisticated stepsize correction, but this
  //       simple scheme seems to work as well as anything else.
  static double vtemparr[3];  static ThreeVector vtemp(vtemparr);
  ThreeVector** htemp; //  Temp. pointer used for swapping h and torque
  double orig_energy,orig_stepsize=StepSize;
  double predicted_change,actual_change;  // Change in energy
  double error;
  int i,k;

  // Save current m, h and energy
  m.Swap(m0);
  htemp=h; h=h0; h0=htemp;
  orig_energy=energy;

  while(1) {
    hValid=0; predicted_change=0.0;
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
      vtemp=m0[i][k];
      m[i][k].FastAdd(1.-SQ(StepSize)*torque[i][k].MagSq()/2.,vtemp,
		      StepSize,torque[i][k]);
      m[i][k].Scale(1);
      vtemp-=m[i][k];
      predicted_change+=(vtemp*h0[i][k])*m[i][k].GetMoment();
      // Should the RHS by multiplied by m[i][k].GetMoment() ??? /**/
      // If I leave it off, I get occasional "Step too small" errors.
      // -mjd, 5-Dec-1996.
    }
    predicted_change*=-1; // Torque is -dE/dM
    hUpdate();  // m, h now hold new position
    CalculateEnergy();
    actual_change=energy-orig_energy;
    error=fabs(actual_change-predicted_change)/((AllowedSpinError*Nx)*Nz);
    if(actual_change<0 && error<1.0) break; // Good step
    StepSize*=CutRatio; // Bad step
    if((StepSize)<SQRT_EPSILON) {  // Safety
      PlainWarning("Warning in Grid2D::Euler: "
		   "Step (%g) too small",StepSize);
      StepSize=0; next_stepsize=orig_stepsize;
      m.Swap(m0); htemp=h; h=h0; h0=htemp; // Undo changes
      energy=orig_energy;
      return;
    }
  }
  next_stepsize=StepSize*(1.+(IncreaseRatio-1.)*(1.-error));

  htemp=torque; torque=torque0; torque0=htemp; // Save old torque and
  CalculateTorque(m,h,torque);                 // calculate new torque
  torqueValid=1;
}

void Grid2D::StepPredict2_5(double &next_stepsize)
{ // Takes one step using a 2.5 order predictor, with stepsize
  // control.  Assumes m, h, torque & energy are current coming in, and
  // leaves them current going out.  Saves entry values of m, h, and
  // torque in m0, h0, and torque0. m1, h1, and torque1 are used for
  // temporary workspace.
  //   The attempted step size is determined by the value of
  // *this->StepSize on entry.  StepSize may be adjusted by this routine;
  // the value of StepSize on exit is the step size actually used.
  // The export next_stepsize should be used by the calling routine (on
  // the next pass) to set *this->StepSize _before_ calling this routine.
  //   If ODEIterCount<1, this routine just calls StepEuler.  Otherwise,
  // it assumes m0, torque0 and StepSize0 have valid values from the
  // previous step.  The controlling program *must* call ODEReset() to
  // reset ODEIterCount on each new integration segment (i.e., when
  // m0, torque0 or StepSize0 are invalid).

  const double CutRatio=0.5;        // Determined by trial & error
  const double IncreaseRatio=1.1;   //             ""

  static double vtemparr[3];  static ThreeVector vtemp(vtemparr);

  if(ODEIterCount<1) { // Insufficient past history to do multistep
    StepEuler(next_stepsize);
    return;
  }

  ThreeVector** htemp; //  Temp. pointer used for swapping h and torque
  int i,k;
  double actual_energy_change;

  double orig_stepsize=StepSize;
  double orig_energy=energy;
  double relstep,tbasecoef,t0coef,tcoef,t1coef;

  while(1) {
    relstep=StepSize/StepSize0;

    // First estimate m1 & h1 in order to estimate torque1
    t0coef= -0.5*relstep*StepSize; tcoef=StepSize*(1+0.5*relstep);
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
      //      h1[i][k].FastAdd((1+relstep),h[i][k],-relstep,h0[i][k]);
#ifdef OLDCODE
      m1[i][k].FastAdd(1.-SQ(StepSize)*torque[i][k].MagSq()/2.,m[i][k],
      		       StepSize,torque[i][k]); // Includes correction from
                                               // |m1[i][k]| \in S^2
#else
//      m1[i][k].FastAdd(m[i][k],tcoef,torque[i][k]);
//      m1[i][k].Accum(t0coef,torque0[i][k]);
      vtemp.FastAdd(t0coef,torque0[i][k],tcoef,torque[i][k]);
      m1[i][k].FastAdd(vtemp,1.-vtemp.MagSq()/2.,m[i][k]);
      // Includes correction due to restriction that |m1[i][k]| \in S^2
#endif
      m1[i][k].Scale(1);
    }
    m1.Swap(m);
    htemp=h; h=h1; h1=htemp;
    hValid=0; hUpdate(); // Estimate h1 from m1 & hdemag_0
    m1.Swap(m);
    htemp=h; h=h1; h1=htemp;
    CalculateTorque(m1,h1,torque1);

    // Now calculate m1 using m, torque0, torque, and torque1
    tbasecoef=StepSize/(6.*(relstep+1));
    t0coef= -relstep*relstep;
    tcoef=  (relstep+4)*relstep+3;
    t1coef= 2*relstep+3;
    t0coef*=tbasecoef; tcoef*=tbasecoef; t1coef*=tbasecoef;
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
      m1[i][k].FastAdd(t0coef,torque0[i][k],tcoef,torque[i][k]);
      m1[i][k].Accum(t1coef,torque1[i][k]);
      m1[i][k]+=m[i][k];
      m1[i][k].Scale(1);
    }
    hValid=0;
    m1.Swap(m); // Swap new m with m on entry
    htemp=torque; torque=torque1; torque1=htemp;
    htemp=h; h=h1; h1=htemp; // Save old field
    hUpdate();  // m, h now hold new position
    CalculateEnergy();
    CalculateTorque(m,h,torque);  // Calculate "real" new torque
    torqueValid=1;

    actual_energy_change=energy-orig_energy;
    if(actual_energy_change<0) break; // Good step
    // Else, bad step
    m1.Swap(m); htemp=h; h=h1; h1=htemp; // Undo changes
    htemp=torque; torque=torque1; torque1=htemp;
    energy=orig_energy;
    StepSize*=CutRatio;
    if((StepSize)<SQRT_EPSILON) {  // Safety
      PlainWarning("Warning in Grid2D::StepPredict2_5: "
		   "Step (%g) too small",StepSize);
      StepSize=0; next_stepsize=orig_stepsize;
      return;
    }
  }
  // Save entry values of m, h and torque
  m1.Swap(m0); htemp=h0; h0=h1; h1=htemp;
  htemp=torque0; torque0=torque1; torque1=htemp;

  next_stepsize=StepSize*(1.+(IncreaseRatio-1.)*1.0);
}

void Grid2D::IncRungeKutta4(Array2D<MagElt> &m_in,
			    ThreeVector** &h_in,
			    double step,
			    Array2D<MagElt> &m_out)
{ // Takes one 4th order Runge-Kutta step, w/o stepsize control.
  // Used as base stepper for Grid2D::StepRungeKutta4()
  // m_in and h_in are imports; results stored in m_out
  // Overwrites m, h, & torque (used for temporary storage)
  // NOTE: m_out *must* be distinct from m_in, and both must be distinct
  //       from m;  Also, h_in must be distinct from h.
  int i,k;
  
  CalculateTorque(m_in,h_in,torque); // Torque at starting location

  // Test step 1
  hValid=0;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m_out[i][k].FastAdd(m_in[i][k],step/6.0,torque[i][k]);
    m[i][k].FastAdd(m_in[i][k],step/2.0,torque[i][k]);
    m[i][k].Scale(1);  // Might want to delay rescaling, but if we do,
                       // will hUpdate() & CalculateTorque() work properly?
  }
#ifdef RUNGE_KUTTA_FAST
  hFastUpdate();  // m, h now hold first test location
#else
  hUpdate();      // m, h now hold first test location
#endif

  // Test step 2
  CalculateTorque(m,h,torque);
  hValid=0;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m_out[i][k].Accum(step/3.0,torque[i][k]);
    m[i][k].FastAdd(m_in[i][k],step/2.0,torque[i][k]);
    m[i][k].Scale(1);  // Ditto rescaling comment.
  }
#ifdef RUNGE_KUTTA_FAST
  hFastUpdate();  // m, h now hold second test location
#else
  hUpdate();      // m, h now hold second test location
#endif

  // Test step 3
  CalculateTorque(m,h,torque);
  hValid=0;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m_out[i][k].Accum(step/3.0,torque[i][k]);
    m[i][k].FastAdd(m_in[i][k],step,torque[i][k]);
    m[i][k].Scale(1);  // Ditto rescaling comment.
  }
#ifdef RUNGE_KUTTA_FAST
  hFastUpdate();  // m, h now hold third test location
#else
  hUpdate();      // m, h now hold third test location
#endif

  // Test step 4
  CalculateTorque(m,h,torque);
  hValid=0;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m_out[i][k].Accum(step/6.0,torque[i][k]);
    m_out[i][k].Scale(1);
  }
}

void Grid2D::StepRungeKutta4(double &next_stepsize)
{ // Takes one 4th order Runge-Kutta step, with adaptive stepsize
  // control.  Assumes m, h, torque & energy are current coming in, and
  // leaves them current going out.
  // The attempted step size is determined by the value of *this->StepSize
  // on entry.  StepSize may be adjusted by this routine; the value of
  // StepSize on exit is the step size actually used.  The export
  // next_stepsize should be used by the calling routine (on the next pass)
  // to set *this->StepSize _before_ calling this routine.

  const double SafetyFactor=0.8;  // Set by Trial & Error
  const double HeadRoom=0.9;      // Ditto
  static double vtemparr[3];  static ThreeVector vtemp(vtemparr);
  ThreeVector** htemp; //  Temp. pointer used for swapping h
  double halfstep,fullstep;
  double oldenergy,error,dtemp,orig_stepsize=StepSize;
  int i,k;

  // Save current m, h and energy
  m.Swap(m0); htemp=h; h=h0; h0=htemp;
  oldenergy=energy;

  // Do one valid step
  next_stepsize=StepSize;
  while(1) {
    if((next_stepsize)<SQRT_EPSILON) {  // Safety
      PlainWarning("Grid2D::StepRungeKutta4: "
		  "Step (%g) too small",next_stepsize);
      StepSize=0; next_stepsize=orig_stepsize;
      m.Swap(m0); htemp=h; h=h0; h0=htemp; energy=oldenergy; // Undo changes
      return;
    }
    halfstep=(fullstep=next_stepsize)/2.0;
    // Do first half step, saving result in m1
    IncRungeKutta4(m0,h0,halfstep,m1);
    // Prepare for second half step
    m1.Swap(m); hValid=0; hUpdate();
    m.Swap(m1); htemp=h; h=h1; h1=htemp;
    // Do second half step, saving result in m2
    IncRungeKutta4(m1,h1,halfstep,m2);

    // Do full step, saving result in m1
    IncRungeKutta4(m0,h0,fullstep,m1);
    StepSize=fullstep;

    // Determine error estimate
    error=0; hValid=0;
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
      vtemp.FastAdd(m2[i][k],-1.0,m1[i][k]);
      if((dtemp=vtemp.MagSq())>error) error=dtemp;
#ifdef OLD_CODE
      m[i][k].FastAdd(m2[i][k],1./15.,vtemp);  // <== Here is final m!
#endif  // On 1st standard problem, straight m2 converges in fewer steps.
    }
#ifndef OLD_CODE
    m.Swap(m2);   // <== Here is final m!
#endif
    error=MAX(sqrt(error),AllowedSpinError/1024.);
    // The "1024." above is arbitrary, used to insure error>0.
    // This also keeps next_stepsize from increasing by more than
    // a factor of 4 in one step.

    // Adjust step size based on error estimate & energy, and
    // terminate "one valid step" loop if appropriate.
    next_stepsize=fullstep*pow(AllowedSpinError/error,0.2)*SafetyFactor;
    if(error>AllowedSpinError) continue; // Bad step; cut and retry
    hUpdate(); CalculateEnergy(); // Else check that energy is decreasing
    if(energy<oldenergy) break;   // Good step!
    next_stepsize=fullstep/2.0; // Bad step;  Cut and retry
  }
  next_stepsize*=HeadRoom;
  CalculateTorque(m,h,torque); torqueValid=1;
}

double Grid2D::StepODE(int& errorcode)
{ // Iterates one step, updating m, h, and energy.
  // Returns max torque value.
  // Will update h as needed.  Will update energy & torque only if
  // hValid is FALSE or ODEIterCount<1.  ***THIS SHOULD BE CONSIDERED***
  //                                     ***   A BUG\b\b\bFEATURE!   ***
  // errorcode>0 on error, =0 on success, <0 if successful with warnings
#define PERTURB_RETRIES 3  // Number of perturbation retries before
                          /// failing
  int perturb_count=0;
  static double next_stepsize=-1;
  double maxtorque;

  errorcode=0;
  StepSize0=StepSize; // Store size of previous step

  // Update h, energy and torque, as necessary
  if(!hValid) { hUpdate();  CalculateEnergy(); }
  if(!torqueValid) {  // NOTE: hUpdate marks toqueValid false
    CalculateTorque(m,h,torque); torqueValid=1;
    maxtorque=GetMaxTorque();
    if(next_stepsize*maxtorque>MaxTorqueStep) {
      next_stepsize=MaxTorqueStep/maxtorque;
    }
  }

  // Take one step
  while(1) {
    if(next_stepsize>0) StepSize=next_stepsize;
#ifdef EULER_ODE               // 1st order Euler
    StepEuler(next_stepsize);
#elif defined(RUNGE_KUTTA_ODE) // 4th order Runge-Kutta
    StepRungeKutta4(next_stepsize);
#elif defined(PREDICTOR_2_5_ODE)
    StepPredict2_5(next_stepsize);
#else
#error Unimplemented ODE solver request
#endif
    if(StepSize>0) break;  // Valid step found
    if(perturb_count>=PERTURB_RETRIES) break; // Give up
    // Otherwise, no valid step found.  Perturb and retry
    MessageLocker::Append("Warning in Grid2D::StepODE: "
			  "Unable to find a valid step,"
			  "  Will jiggle values and retry.\n");
    perturb_count++;
    Perturb(PERTURBATION_SIZE); hUpdate();  CalculateEnergy();
    CalculateTorque(m,h,torque); torqueValid=1;
    maxtorque=GetMaxTorque();
    if(next_stepsize*maxtorque>MaxTorqueStep)
      next_stepsize=MaxTorqueStep/maxtorque;
  }

  // Adjust next_stepsize as required by MaxTorqueStep
  maxtorque=GetMaxTorque();
  if(next_stepsize*maxtorque>MaxTorqueStep) {
    next_stepsize=MaxTorqueStep/maxtorque;
  }

  // Check for errors, and increment ODEIterCount if applicable
  if(StepSize<=0) {
    errorcode=1;
    MessageLocker::Append("ERROR in Grid2D::StepODE: "
			  "Unable to find a valid step,"
			  "  Giving up.\n");
  } else {
    ODEIterCount++;
    errorcode = -perturb_count;
  }

  return maxtorque;
#undef PERTURB_RETRIES
}

void Grid2D::GetNearestGridNode(double x,double z,int &i,int &k)
{
#ifdef CODE_CHECKS
  if(cellsize<SQ(EPSILON))
    PlainError(1,"Error in Grid2D::GetNearestGridNode: "
	     "Cellsize (=%g) is too small or not initialized",cellsize);
#endif // CODE_CHECKS
  i=int(ceil(x/cellsize));  k=int(ceil(z/cellsize));
  if(i<0) i=0; else if(i>=Nx) i=Nx-1;
  if(k<0) k=0; else if(k>=Nz) k=Nz-1;
}

void Grid2D::GetRealCoords(int i,int k,double &x,double & z)
{ // Node points are in cell centers
  if(Nx<1) x=0.0; else x=xmax*double(i)/double(Nx);
  if(Nz<1) z=0.0; else z=zmax*double(k)/double(Nz);
  x+=cellsize/2.; z+=cellsize/2.;
}

double Grid2D::GetMaxNeighborAngle(void)
{ // Returns maximum angle (in radians) between
  // neighboring spin directions in the grid.
  int i,k;
  double dot,mindot;
  mindot=1.0;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    // Note: Spins with zero-moment *should* have
    //  no neighbors, and therefore return dot=1.0
    dot=m[i][k].GetMinNeighborDot();
    if(dot<mindot) mindot=dot;
  }
  if(fabs(mindot)>1) {
    fprintf(stderr,"ERROR: Mindot=%15.12f\n",mindot);
    PlainError(1,"Error in Grid2D::GetMaxNeighborAngle: %s",ErrProgramming);
  }
  return acos(mindot);
}

Vec3D Grid2D::GetAveMag()
{
  int i,k;
  Vec3D AveMag(0.,0.,0.);
#ifndef UNIT_MAG
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
    AveMag.Accum(m[i][k].GetMoment(),m[i][k]);
#else
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) AveMag+=m[i][k];
#endif
  AveMag*=(1.0/moment_sum);
  return AveMag;
}

void Grid2D::GetAverageMagnetization(ThreeVector& AveMag)
{
  Convert(AveMag,GetAveMag());
}

void Grid2D::GetMag(double** arr)
{ // Fills arr with magnetizations.  The calling routine must
  // make sure arr is big enough (cf. Grid2D::GetDimens()).
  int i,k;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
    m[i][k].DumpMag(arr[i]+k*VECDIM);
}

void Grid2D::Dump(CHAR *filename,CHAR *note)
{
  int i,k;
  FILE *fptr;
  double Bapparr[3]; ThreeVector Bapp(Bapparr);
  hApplied->GetLocalBField(0,0,Bapp);

  if((fptr=Omf_FOpen(filename,"wb"))==NULL) {
    fprintf(stderr,"FATAL ERROR: Unable to open output file %s\n",
	    filename);
    exit(1);
  }

  double maxtorque=GetMaxTorque();
  double e_exch,e_anis,e_demag,e_zeeman,e_total;
  GetEnergies(e_exch,e_anis,e_demag,e_zeeman,e_total);

  if(note!=NULL) fprintf(fptr,"%s\n\n",note);
  fprintf(fptr,
	  "\nMicromagnetic grid parameters-------\n"
	  " Grid size                     %3dx%d\n"
	  " Demag field routine        demag = %s\n"     
	  " Saturation magnetization      Ms = %g (A/m)\n"
	  " Anisotropy constant           K1 = %g (J/m^3)\n"
	  " Exchange stiffness             A = %g (J/m)\n"
	  " Applied field               Bapp = <%g,%g,%g> (T)\n"
	  " Mesh cell edge length       mesh = %g (m)\n"
	  " Final step size         StepSize = %g\n\n",
	  Nx,Nz,ExternalDemag->name,Ms,K1,exchange_stiffness,
	  Bapp[0],Bapp[1],Bapp[2],
	  mesh_size,StepSize);
  fprintf(fptr,"Micromagnetic modeling results-------\n");
  fprintf(fptr,
	  "Energy:  Total = %10.6f\n"
	  "          Exch = %10.6f\n"
	  "          Anis = %10.6f\n"
	  "         Demag = %10.6f\n"
	  "        Zeeman = %10.6f\n",
	  e_total,e_exch,e_anis,e_demag,e_zeeman);
  fprintf(fptr,"Max torque: %10.7g\n",maxtorque);
#ifndef UNIT_MAG
  double mom;
#endif
  for(i=0;i<Nx;i++) {
    fprintf(fptr,"\nRow %d:\n",i);
    for(k=0;k<Nz;k++) {
#ifndef UNIT_MAG
      mom=m[i][k].GetMoment();
      fprintf(fptr,"m[%2d][%2d]=(%7.4f,%7.4f,%7.4f)"
	      "\tt[%2d][%2d]=(%10.7f,%10.7f,%10.7f)\n",
	      i,k,mom*m[i][k][0],mom*m[i][k][1],mom*m[i][k][2],
	      i,k,torque[i][k][0],torque[i][k][1],torque[i][k][2]);
#else
      fprintf(fptr,"m[%2d][%2d]=(%7.4f,%7.4f,%7.4f)"
	      "\tt[%2d][%2d]=(%10.7f,%10.7f,%10.7f)\n",
	      i,k,m[i][k][0],m[i][k][1],m[i][k][2],
	      i,k,torque[i][k][0],torque[i][k][1],torque[i][k][2]);
#endif // UNIT_MAG
    }
  }

  fprintf(fptr,"\n\n");
  fclose(fptr);
}

// Binary dump of (reduced) magnetization (VecFil format)
void Grid2D::BinDumpm(CHAR *filename,CHAR *note)
{
  int i;
  FileVector fv;

  fv.SetNote((BYTE *)note);

  if(fv.WriteHeader(filename,Nx,1,Nz,VECDIM)) {
    PlainWarning("Warning in Grid2D::BinDumpm: %s",ErrBadFile);
    return;
  }

  int k;
  double *temparr= new double[VECDIM*Nz];
  for(i=0;i<Nx;i++) {
    for(k=0;k<Nz;k++) m[i][k].DumpMag(temparr+k*VECDIM);
    if(fv.WriteVecs(Nz,temparr)) {
      PlainWarning("Warning in Grid2D::BinDumpm: %s",ErrBadFile);
      break;
    }
  }
  delete temparr;
  fv.Close();
}

// Binary dump of (reduced) field (VecFil format)
void Grid2D::BinDumph(CHAR *filename,CHAR *note)
{
  int i;
  FileVector fv;

  if(!hValid) { hUpdate(); CalculateEnergy(); }

  fv.SetNote((BYTE *)note);

  if(fv.WriteHeader(filename,Nx,1,Nz,VECDIM)) {
    PlainWarning("Warning in Grid2D::BinDumph: %s",ErrBadFile);
    return;
  }

  for(i=0;i<Nx;i++) {
    if(fv.WriteVecs(Nz,&(h[i][0][0]))) { // NOTE: &(h[i][0][0])     /**/
                                         // is a *BAD* thing        /**/
      PlainWarning("Warning in Grid2D::BinDumph: %s",ErrBadFile);
      break;
    }
  }
  fv.Close();
}

void WriteGrid2Dm(double** dat,void* m,int xdim,int ydim,int xyz)
{
  int i,k;
  int component=0,sign=1;
  switch(xyz) // Adjust for differences in component ordering
    {
    case 0: component=0;  sign= 1; break;
    case 1: component=2;  sign= 1; break;
    case 2: component=1;  sign=-1; break;  // Actually, this sign shouldn't
	    // matter because it gets reversed back in FillGrid2Dh()
    default: PlainError(1,"Error in WriteGrid2Dm: Illegal component request: %d",xyz);
	    break;
    }
#ifndef UNIT_MAG
  MagElt* mptr;
  for(i=0;i<xdim;i++) for(k=0,mptr=((MagElt**)m)[i];k<ydim;k++,mptr++)
    dat[i][k]=sign*(mptr->GetMoment())*((*mptr)[component]);
#else
  for(i=0;i<xdim;i++) for(k=0;k<ydim;k++)
    dat[i][k]=sign*((MagElt**)m)[i][k][component];
#endif

#ifdef EDGE_ADJUST
/**/ // Soften demag on edges
    for(k=0;k<ydim;k++) { 
      dat[0][k]*=EDGE_ADJUST; dat[xdim-1][k]*=EDGE_ADJUST;
    }
    for(i=1;i<xdim-1;i++) {
      dat[i][0]*=EDGE_ADJUST; dat[i][ydim-1]*=EDGE_ADJUST;
    }
#endif // EDGE_ADJUST
}

void FillGrid2Dh(double** dat,void* h,int xdim,int ydim,int xyz)
{
  int i,k;
  int component=0,sign=1;
  switch(xyz) // Adjust for differences in component ordering
    {
    case 0: component=0; sign= 1; break;
    case 1: component=2; sign= 1; break;
    case 2: component=1;  sign=-1; break;  // Actually, this sign shouldn't
	    // matter because it gets reversed back in FillGrid2Dh()
    default: PlainError(1,"Error in FillGrid2Dm: Illegal component request: %d",xyz);
	    break;
    }

#ifdef EDGE_ADJUST
/**/ // Soften demag on edges
    for(k=0;k<ydim;k++) { 
      dat[0][k]*=EDGE_ADJUST; dat[xdim-1][k]*=EDGE_ADJUST;
    }
    for(i=1;i<xdim-1;i++) {
      dat[i][0]*=EDGE_ADJUST; dat[i][ydim-1]*=EDGE_ADJUST;
    }
#endif // EDGE_ADJUST

  for(i=0;i<xdim;i++) for(k=0;k<ydim;k++)
    ((ThreeVector**)h)[i][k][component]=sign*dat[i][k];
}

