# FILE: connection.tcl
#
# An object which manages asynchronous communication over a net_link.
#
# Query, expects reply
# Tell, just sends message
#
# The Tcl event loop must be running for this event-driven object to function.
#

em_class net_connection {

    # The link which manages lower level of communications.
    private variable link

    # The socket we hand off to our link
    public variable socket 

    # The hostname we hand off to our link
    public variable hostname 

    # The port we hand off to our link
    public variable port 
    
    # The protocol used to handle queries received over this connection.
    public variable protocol
    private variable myProtocol

    # State of query handling protocol state machine -- move this into protocol
    private variable state

    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler delete
        }
        set logHandler [em_eventHandler new $log delete "$this delete"]
        if {$imadelog} {
            $oldlog delete
        }
        set oldlog $log
    }
    private variable oldlog = em_nop
    private variable logHandler
    private variable imadelog = 0

    # Received message
    private variable message

    # query id
    private variable id = 0

    # Ready to send/receive?
    private variable ready = 0

    # Array of timeout events
    private variable toe
    public variable timeout = 60000

    # My binding tags
    private variable bindtags
    private variable myEventHandlers = {}

    # Constructor is a non-blocking operation.  
    constructor { args } {
        eval $this configure $args
        if {![info exists log]} {
            # Provide a default log
            $this configure -log [em_log new -title "Connection $this Log"]
            set imadelog 1
        }
        # Enable event generation
        set bindtags [list $this $class]
        if {[info exists socket]} {
            # Assume a server
            if {![info exists protocol]} {
                error "Server connections need a query-handling protocol"
            }
            set myProtocol [$protocol Clone $this]
            lappend myEventHandlers [em_eventHandler new $myProtocol delete \
                    [list $this delete]]
            set link [net_link new -socket $socket -log $log]
            lappend myEventHandlers [em_eventHandler new $link Ready \
                    [list $this ServerLinkReady]]
        } elseif {[info exists hostname] && [info exists port]} {
            # Assume a client
            set link [net_link new -hostname $hostname -port $port -log $log]
            lappend myEventHandlers [em_eventHandler new $link Ready \
                    [list $this LinkReady]]
        } else {
            error "Missing required options: -hostname and -port or -socket"
        }
        lappend myEventHandlers [em_eventHandler new $link delete \
                [list $this delete]]
    }

    method ServerLinkReady {} {
        $log Append "Put [concat tell [$myProtocol cget -name]]"
        $link Put [concat tell [$myProtocol cget -name]]
        $this LinkReady
    }

    method LinkReady {} {
        lappend myEventHandlers [em_eventHandler new $link Readable \
                [list $this Receive]]
        set hostname [$link cget -hostname]
        set port [$link cget -port]
        set toe [em_array new]
        set state start
        set ready 1
        em_eventHandler Generate Ready $bindtags
    }

    # Every line of data received over the link gets handled by this
    # method.  It has to parse the message and direct it to its 
    # destination.
    method Receive {} {
        # NOTE: The function of the net_link is to transmit and receive
        # strings.  The net_connection wants to interpret these strings
        # as Tcl lists.  Not all string are a valid Tcl list, so we have
        # to be sure the line of input we receive is tested, or somehow
        # interpreted as a valid Tcl list.  I would like to use the
        # split and join commands to translate back and forth between
        # lists and strings but I can't because contrary to their 
        # documentation, they are *not* inverses of each other.
        # For example:
        # % split [join {1 {2 3} {4 5 6}}]
        # 1 2 3 4 5 6
        # The list structure can be completely changed. Grrrrrrr..
        # It seems the best we can do is use the Tcl list commands to
        # at least test the the string we receive has a valid interpretation
        # as a Tcl list, and log an error and bail if it does not.

        # Get the string
        set rawmessage [$link Get]
        $log Append "Received raw message: $rawmessage"
        # Test for list structure 
        # The 'eval' is required to expand the string $rawmessage into
        # whatever words (list elements) it contains.
        if {[catch {eval list $rawmessage} msg]} {
            $log Append "Message not a Tcl list: $rawmessage"
            return
        }
        if {![llength $msg]} {
            # Ignore empty messages
            return
        }
        switch -- [lindex $msg 0] {
            reply {
                if {[llength $msg] == 1} {
                    $log Append "Received bad message: $msg"
                    return
                }
                set rid [lindex $msg 1]
                if {![catch {$toe lookup $rid} ev]} {
                    after cancel $ev
                    $toe unset $rid
                }
                set message [lrange $msg 2 end]
                em_eventHandler Generate Reply$rid $bindtags
            }
            query {
                if {[llength $msg] == 1} {
                    $log Append "Received bad message: $msg"
                    return
                }
                set id [lindex $msg 1]
                set message [lrange $msg 2 end]
                $log Append "Handling query: $message"
                # This doesn't allow for event-driven query handling.
                # Should it?  If so, should it timeout?
		set reply {}   ;# Safety
                foreach {state reply} [$myProtocol Reply $state $message] {}
                $log Append "Put [concat reply $id $reply]"
                $link Put [concat reply $id $reply]
                if {[string match close $state]} {
                    $log Append "Closing connection on remote command"
                    $this delete
                }
            }
            tell {
                set message [lrange $msg 1 end]
                em_eventHandler Generate Readable $bindtags
            }
            default {
                # Log non-conforming messages, but continue.
                $log Append "Received bad message: $msg"
            }
        }
    }

    method Get {} {
        return $message
    }

    method Tell { args } {
        if {!$ready} {
            error "connection not ready"
        }
        $log Append "Put [concat tell $args]"
        $link Put [concat tell $args]
    }

    method Query { args } {
        if {!$ready} {
            error "connection not ready"
        }
        incr id
        $toe set [list $id [after $timeout $this Timeout $id]]
        $log Append "Put [concat query $id $args]"
        $link Put [concat query $id $args]
        return $id
    }

    method Timeout { qid } {
        $log Append "Timeout waiting for reply to query $qid"
        em_eventHandler Generate Timeout$qid $bindtags
    }

    destructor {
        if {[info exists bindtags]} {
            em_eventHandler Generate delete $bindtags
        }
        # delete handlers before deleting event-generating members
        if {[info exists toe]} {
            foreach i [$toe names] {
                after cancel [$toe lookup $i]
                $toe unset $i
            }
        }
        if {[info exists myEventHandlers]} {
            foreach handler $myEventHandlers {
                $handler delete
            }
        }
        if {[info exists myProtocol]} {
            $myProtocol delete
        }
        if {[info exists link]} {
            $link delete
        }
        if {[info exists log]} {
            $log Append "Deleting $this ..."
            $this configure -log em_nop
        }
    }

}

